__doc__ = """

-------------------------------------

Raw Data Loader
###############

"""
__all__ = [
    "load_continuous_data",
    "load_recording",
    "oebin_read",
    "apply_channel_mask",
    "load_ttl_event",
]

from typing import Any, Dict, List, Optional, Sequence, Set, Union

import logging
import os
from ast import literal_eval
from glob import glob

import neo
import numpy as np
import quantities as pq
from tqdm import tqdm

from miv.typing import SignalType, TimestampsType


def apply_channel_mask(signal: np.ndarray, channel_mask: Set[int]):
    """Apply channel mask on the given signal.

    Parameters
    ----------
    signal : np.ndarray
        Shape of the signal is expected to be (num_data_point, num_channels).
    channel_mask : Set[int]

    Returns
    -------
    output signal : SignalType

    Raises
    ------
    IndexError
        Typically raise index error when the dimension of the signal is less than 2.
    AttributeError
        If signal is non numpy array type.

    """

    num_channels = signal.shape[1]
    channel_index_set = set(range(num_channels)) - channel_mask
    channel_index = np.array(np.sort(list(channel_index_set)))
    signal = signal[:, channel_index]
    return signal


def bits_to_voltage(signal: SignalType, channel_info: Sequence[Dict[str, Any]]):
    """
    Convert binary bit data to voltage (microVolts)

    Parameters
    ----------
    signal : SignalType, numpy array
    channel_info : Dict[str, Dict[str, Any]]
        Channel information dictionary. Typically located in `structure.oebin` file.
        Channel information includes bit-volts conversion ration and units (uV or mV).

    Returns
    -------
    signal : numpy array
        Output signal is in microVolts (uV) unit.

    """
    resultant_unit = pq.Quantity(1, "uV")  # Final Unit
    for channel in range(len(channel_info)):
        bit_to_volt_conversion = channel_info[channel]["bit_volts"]
        recorded_unit = pq.Quantity([1], channel_info[channel]["units"])
        unit_conversion = (recorded_unit / resultant_unit).simplified
        signal[:, channel] *= bit_to_volt_conversion * unit_conversion
    return signal


def oebin_read(file_path: str):
    """
    Oebin file reader in dictionary form

    Parameters
    ----------
    file_path : str

    Returns
    -------
    info : Dict[str, any]
        recording information stored in oebin file.
    """
    # TODO: may need fix for multiple continuous data.
    # TODO: may need to include processor name/id
    info = literal_eval(open(file_path).read())
    return info


def load_ttl_event(
    folder: str,
    return_sample_numbers: bool = False,
):
    """
    Loads TTL event data recorded by Open Ephys as numpy arrays.

    `Reference: OpenEphys TTL data structure <https://open-ephys.github.io/gui-docs/User-Manual/Recording-data/Binary-format.html#events>`_

    The path should contain:

    - states.npy: N 16-bit integers, indicating ON/OFF (channel number)
    - sample_numbers.npy: N 64-bit integers, sample number during acquisition.
    - timestamps.npy: N 64-bit floats, global timestamps
    - full_words.npy: N 64-bit integer, TTL word of current state of all lines.

    Extra data are retrieved from:

    - structure.oebin: number of channels and sampling rate.

    Parameters
    ----------
    folder: str
        Folder containing the subfolder 'experiment1'.
    return_sample_numbers: bool
        If set to true, also return sample_numbers that can be used to re-calculate
        synchronization between time series. (default=False)

    Returns
    -------
    states : np.ndarray
        Numpy integer array, indicating ON/OFF state. (+- channel number)
    full_words : np.ndarray
        Numpy integer array, consisting current state of all lines.
    timestamps : TimestampsType
        Numpy float array. Global timestamps in seconds. Relative to start
        of the Record Node's main data stream.
    sampling_rate: float
        Recorded sampling rate
    initial_state: int
        Initial TTL state across lines.
    sample_numbers: Optional[np.ndarray]
        Return if `return_sample_numbers` is true. Return array of sample numbers that
        records sampled clock count. Typically used to synchronize time array.

    Raises
    ------
    AssertionError
        No events recorded in data.

    """

    # Check TTL event recorded
    info_file: str = os.path.join(folder, "structure.oebin")
    info: Dict[str, Any] = oebin_read(info_file)
    version = info["GUI version"]
    assert "events" in info.keys(), "No events recorded (TTL)."
    ttl_info = [
        data
        for data in info["events"]
        if "TTL Input" in data["channel_name"] or "TTL events" in data["channel_name"]
    ]
    assert len(ttl_info) > 0, "No events recorded (TTL)."
    assert (
        len(ttl_info) == 1
    ), "Multiple TTL input is found, which is not supported yet. (TODO)"
    ttl_info = ttl_info[0]

    # Data Structure (OpenEphys Structure)
    v_major, v_minor, v_sub = map(int, version.split("."))
    if v_major == 0 and v_minor <= 5 and v_sub == 4:  # Legacy file name before 0.6.0
        file_states = "channel_states.npy"
        file_timestamps = "timestamps.npy"
        file_sample_numbers = "channels.npy"
        file_full_words = "full_words.npy"
    elif v_major == 0 and v_minor <= 5:  # Legacy file name before 0.6.0
        file_states = "states.npy"
        file_timestamps = "synchronized_timestamps.npy"
        file_sample_numbers = "timestamps.npy"
        file_full_words = "full_words.npy"
    else:
        file_states = "states.npy"
        file_timestamps = "timestamps.npy"
        file_sample_numbers = "sample_numbers.npy"
        file_full_words = "full_words.npy"
    file_path = os.path.join(folder, "events", ttl_info["folder_name"])

    states = np.load(os.path.join(file_path, file_states)).astype(np.int16)
    sample_numbers = np.load(os.path.join(file_path, file_sample_numbers)).astype(
        np.int64
    )
    timestamps = np.load(os.path.join(file_path, file_timestamps)).astype(np.float64)
    full_words = np.load(os.path.join(file_path, file_full_words)).astype(np.int64)

    # Load from structure.oebin file
    sampling_rate: float = ttl_info["sample_rate"]
    initial_state: int = ttl_info["initial_state"] if "initial_state" in ttl_info else 0

    if return_sample_numbers:
        return (
            states,
            full_words,
            timestamps,
            sampling_rate,
            initial_state,
            sample_numbers,
        )
    else:
        return states, full_words, timestamps, sampling_rate, initial_state


def load_recording(
    folder: str,
    channel_mask: Optional[Set[int]] = None,
    start_at_zero: bool = True,
    num_fragments: int = 1,
    start_index: Optional[int] = None,
    end_index: Optional[int] = None,
    dtype: np.dtype = np.float32,
    verbose: bool = False,
    progress_bar: bool = False,
):
    """
    Loads data recorded by Open Ephys in Binary format as numpy memmap.
    The path should contain

    - continuous/<processor name>/continuous.dat: signal (cannot have multiple file)
    - continuous/<processor name>/timestamps.npy: timestamps
    - structure.oebin: number of channels and sampling rate.

    Parameters
    ----------
    folder: str
        folder containing at least the subfolder 'experiment1'.
    channel_mask: Set[int], optional
        Channel index list to ignore in import (default=None)
    start_at_zero : bool
        If True, the timestamps is adjusted to start at zero.
        Note, recorded timestamps might not start at zero for some reason.
    num_fragments : int
        Instead of loading entire data at once, split the data into `num_fragment`
        number of subdata to process separately. (default=1)
    start_index : Optional[int]
        Start index of the fragments. It is useful when you want to submit MPI processing.
        For example, one can submit num_fragments=10, start_index=3 to process 3-9 fragments.
        (Zero-indexing)
    end_index : Optional[int]
        End index of the fragments. It is useful when you want to submit MPI processing.
        For example, one can submit num_fragments=10, end_index=5 to process 0-5 fragments.
        (Zero-indexing)
    dtype: np.dtype
        If None, skip data-type conversion. If the filesize is too large, it is advisable
        to keep `dtype=None` and convert slice by slice. (default=float32)
    verbose: bool
        Include logging.info messages for intermediate progresses. (default=False)

    Returns
    -------
    signal : SignalType, neo.core.AnalogSignal
    timestamps : TimestampsType
    sampling_rate : float

    Raises
    ------
    AssertionError
        If more than one "continuous.dat" file exist in the directory.

    """

    file_path: List[str] = glob(
        os.path.join(folder, "**", "continuous.dat"), recursive=True
    )
    assert (
        len(file_path) == 1
    ), f"There should be only one 'continuous.dat' file. (There exists {file_path})"

    # load structure information dictionary
    info_file: str = os.path.join(folder, "structure.oebin")
    info: Dict[str, Any] = oebin_read(info_file)
    num_channels: int = info["continuous"][0]["num_channels"]
    sampling_rate: float = float(info["continuous"][0]["sample_rate"])
    # channel_info: Dict[str, Any] = info["continuous"][0]["channels"]

    # TODO: maybe need to support multiple continuous.dat files in the future
    signal, timestamps = load_continuous_data(file_path[0], num_channels, sampling_rate)
    fragmented_signal = np.array_split(signal, num_fragments, axis=0)[
        start_index:end_index
    ]
    fragmented_timestamps = np.array_split(timestamps, num_fragments)[
        start_index:end_index
    ]
    for frag_id in tqdm(range(num_fragments), disable=not progress_bar):
        _signal = fragmented_signal[frag_id]
        _timestamps = fragmented_timestamps[frag_id]
        if _signal.shape[0] > 6.0e6:
            logging.warn(
                "The recording data is too long. Consider using more num_fragments."
            )
        _signal = _signal.astype(dtype)
        if verbose:
            logging.info("Cast done")

        # To Voltage
        _signal = bits_to_voltage(_signal, info["continuous"][0]["channels"])

        if channel_mask:
            _signal = apply_channel_mask(_signal, channel_mask)

        # Adjust timestamps to start from zero
        if start_at_zero and not np.isclose(_timestamps[0], 0.0):
            _timestamps -= _timestamps[0]

        yield _signal, _timestamps, sampling_rate


def load_continuous_data(
    data_path: str,
    num_channels: int,
    sampling_rate: float,
    timestamps_path: Optional[str] = None,
    _recorded_dtype: Union[np.dtype, str] = "int16",
):
    """
    Load single continous data file and return timestamps and raw data in numpy array.
    Typical `data_path` from OpenEphys has a name `continuous.dat`.

    .. note::
        The output data is raw-data without unit conversion. In order to convert the unit
        to voltage, you need to multiply by `bit_volts` conversion ratio. This ratio and
        units are typially saved in `structure.oebin` file.

    Parameters
    ----------
    data_path : str
        continuous.dat file path from Open_Ethys recording.
    num_channels : int
        number of recording channels recorded. Note, this method will not throw an error
        if you don't provide the correct number of channels.
    sampling_rate : float
        data sampling rate.
    timestamps_path : Optional[str]
        If None, first check if the file "timestamps.npy" exists on the same directory.
        If the file doesn't exist, we deduce the timestamps based on the sampling rate
        and the length of the data.
    _recorded_dtype: Union[np.dtype, str]
        Recorded data type. (default="int16")

    Returns
    -------
    raw_data: SignalType, numpy array
    timestamps: TimestampsType, numpy array

    Raises
    ------
    FileNotFoundError
        If data_path is invalid.
    ValueError
        If the error message shows the array cannot be reshaped due to shape,
        make sure the num_channels is set accurately.

    """

    # Read raw data signal
    raw_data: np.ndarray = np.memmap(
        data_path, dtype=_recorded_dtype, mode="r", order="C"
    )
    length = raw_data.size // num_channels
    raw_data = raw_data.reshape(length, num_channels)

    # Get timestamps_path
    if timestamps_path is None:
        dirname = os.path.dirname(data_path)
        timestamps_path = os.path.join(dirname, "timestamps.npy")

    # Get timestamps
    if os.path.exists(timestamps_path):
        timestamps = np.asarray(
            np.load(timestamps_path)
        )  # TODO: check if npy file includes dtype. else, add "dtype=np.float32"
        # timestamps /= float(sampling_rate)  # TODO(beyond 0.3.0) As of OpenEphys 0.6, this line is no-longer necessary.
    else:  # If timestamps_path doesn't exist, deduce the stamps
        timestamps = np.array(range(0, length)) / sampling_rate

    return raw_data, timestamps



__all__ = ["GridMEA"]

from typing import Tuple

import matplotlib
import numpy as np

from miv.mea.protocol import MEAGeometryProtocol


class GridMEA:
    """
    A class representing a grid-based multi-electrode array (MEA).

    Example
    -------
    For example, you could create an instance of the GridMEA class like this:
    literal blocks::

        nrow = 10
        ncol = 20
        xid = np.arange(nrow * ncol) % ncol
        yid = np.arange(nrow * ncol) // ncol
        mea = GridMEA(nrow, ncol, xid, yid)

    Attributes
    ----------
    nrow: int
        The number of rows in the grid.
    ncol: int
        The number of columns in the grid.
    xid: np.ndarray
        A NumPy array containing the x-coordinates of the electrodes in the grid.
    yid: np.ndarray
        A NumPy array containing the y-coordinates of the electrodes in the grid.
    """

    def __init__(self, nrow: int, ncol: int, xid: np.ndarray, yid: np.ndarray):
        self.nrow = nrow
        self.ncol = ncol
        self.xid = xid
        self.yid = yid

    def get_closest_node(self, x: float, y: float) -> int:  # pragma: no cover
        """Given xy coordinate, return closest node idx"""
        raise NotImplementedError

    def get_xy(self, idx: int) -> Tuple[float, float]:  # pragma: no cover
        """Given node index, return xy coordinate"""
        raise NotImplementedError

    def view(self) -> matplotlib.pyplot.Figure:  # pragma: no cover
        """Simplified view of MEA orientation"""
        raise NotImplementedError

    def save(self, path: str) -> None:  # pragma: no cover
        """Export MEA information"""
        raise NotImplementedError

    def load(self, path: str) -> None:  # pragma: no cover
        """Import MEA from external source"""
        raise NotImplementedError



__doc__ = """

One could also take additional steps to fine-adjust the shape of the spikes for clustering and classification tasks

Dynamic time warping
####################

Dynamic time warping is a similarity metrics that accounts for the temporal dilatation and shift in multiple sequences of data [1]_ [wiki]_ [dtaidistance]_.

The tooll is available with the sorting extension: `pip install miv-os[sortingExtension]`.

References & Footnotes
======================

.. [1] John Thomasa, Jing Jina, Justin Dauwelsa, Sydney S. Cashb, and M. Brandon Westoverb. (2018), Clustering of Interictal Spikes by Dynamic Time Warping and Affinity Propagation. IEEE Int Conf Acoust Speech Signal Process (2018 March), https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5842698/pdf/nihms946113.pdf
.. [wiki] https://en.wikipedia.org/wiki/Dynamic_time_warping
.. [dtaidistance] https://dtaidistance.readthedocs.io/en/latest/

.. currentmodule:: miv.signal.similarity

.. autofunction:: dynamic_time_warping_distance_matrix

"""
__all__ = ["dynamic_time_warping_distance_matrix"]
import os
import sys

import numpy as np
from dtaidistance import dtw

from miv.typing import SignalType


def dynamic_time_warping_distance_matrix(
    temporal_sequence: np.ndarray, axis: int = 0, fast: bool = False, **kwargs
):
    """
    Measure dynamic time warping matrix between signals.

    Parameters
    ----------
    temporal_sequence : np.ndarray
        The temporal sequence (time, n_features).
    axis : int
        Axis of time sequence. (default=0)
    fast : bool
        Allow fast computation if `dtaidistance` is compiled with C. (default=False)
        :ref: https://dtaidistance.readthedocs.io/en/latest/usage/installation.html#from-source

    Returns
    -------
    matrix : np.ndarray
        2D matrix with the size (temporal_sequence.shape[axis], temporal_sequence.shape[axis])
        Each element represent the distance between the feature set.
    """
    assert (
        len(temporal_sequence.shape) > 1
    ), f"Sequence must have at least two axes. Provide sequence has a shape {temporal_sequence.shape}."
    if axis != 0:
        temporal_sequence = np.swapaxes(temporal_sequence, axis, 0)

    if fast:
        return dtw.distance_matrix_fast(temporal_sequence, **kwargs)
    else:
        return dtw.distance_matrix(temporal_sequence, **kwargs)



__all__ = ["get_file", "check_file_hash"]

import typing
from typing import Any, Callable, Optional

import hashlib
import logging
import os
import pathlib
import shutil
import tarfile
import urllib
import zipfile

import numpy as np
from six.moves.urllib.request import urlopen
from tqdm import tqdm


def get_file(
    file_url: str,
    directory: str,
    fname: str,
    file_hash: Optional[str] = None,
    archive_format: Optional[str] = "zip",
    cache_dir: str = "datasets",
):
    """
    Downloads a file from a URL if it not already in the directory.
    If the archive_format is provided, the downloaded file will be extracted.

    By default the file at the url `file_url` is downloaded to the
    `datasets/<directory>` with the filename `fname`.

    If the downloaded file is compressed or archived file, the parameter
    `archive_format` can be specified to extract the file after download.
    Currently supports tar (tar.gz, tar.bz) and zip archive format.

    The hash of the downloaded file can be checked by passing `file_hash`
    parameter. We use sha256 hashing algorithm. If `None`, validation process is
    skipped. Note, the hash validation is the only way to skip the download phase;
    if the file is expected to be large and take time to download, make sure
    you add file_hash.

    Parameters
    ----------
    file_url : str
        URL of the file.
    directory : str
        Subdirectory under the cache dir to save the file.
    fname : str
        Name of the downloaded file.
    file_hash : Optional[str]
        The expected hash string of the file (sha256) after download.
        Note, the only way to validate the file is by comparing hash.
        If hash is not give or not valid, the file will be re-downloaded.
        (default=None)
    archive_format :
        Archived format of the downloaded file.  Currently supports `tar`,
        `zip`, and `None`. None will skip the extraction step. (default="zip")
    cache_dir : str
        Location to store files. If not given, file will be installed in "datasets".
        (default="datasets")

    Returns
    -------
    path: str
        Path to the downloaded folder. If archived, path returns the
        unpacked directory.
    """

    # Prepare directory
    data_dir = os.path.join(cache_dir, directory)
    filepath = os.path.join(data_dir, fname)
    os.makedirs(data_dir, exist_ok=True)

    # Check hash
    download = True
    if os.path.exists(filepath) and file_hash is not None:
        # File found; verify integrity if a hash was provided.
        hash_check = check_file_hash(filepath, file_hash)
        if not hash_check:
            logging.info("Re-download the data: hash is invalid")
            download = True  # re-download if hash is different
        else:
            download = False

    # Download
    if download:
        logging.info(f"Downloading: {file_url}")

        error_msg = "URL fetch failure on {}: {} -- {}"
        try:
            try:
                _url_retrieve(file_url, filepath)
            except urllib.error.HTTPError as e:  # pragma: no cover
                raise Exception(error_msg.format(file_url, e.code, e.msg))
            except urllib.error.URLError as e:  # pragma: no cover
                raise Exception(error_msg.format(file_url, e.errno, e.reason))
        except (Exception, RuntimeError, KeyboardInterrupt):  # pragma: no cover
            # Remove residue upon error
            if os.path.exists(filepath):
                os.remove(filepath)
            raise RuntimeError("Interrupted during downloading the file.")

        # Validate download
        if os.path.exists(filepath) and file_hash is not None:
            hash_check = check_file_hash(filepath, file_hash)
            if not hash_check:
                raise ValueError(
                    f"The sha256 file hash does not match "
                    f"the provided value of {file_hash}. If the issue "
                    f"remains, please report this error on GitHub issue."
                )
    else:
        logging.info("Downloading skipped: valid file already exists.")

    if archive_format:
        flag = _extract_archive(filepath, data_dir, archive_format)
        # remove file extension
        if archive_format.endswith((".gz", ".bz")):  # Corner case
            filepath = filepath[:-7]
        else:
            filepath = filepath[:-4]
        assert flag, (
            "File extraction is not completed properly."
            "Please report this error on GitHub issue."
        )

    return filepath


def check_file_hash(filepath: str, file_hash: str, packet_size: int = 65535) -> bool:
    """
    Validates a file using sha256 hash.

    Example
    -------
        >>> check_file_hash(
        ...     filepath='/path/to/file.zip',
        ...     file_hash='e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
        ... )
        True

    Parameters
    ----------
    filepath : str
        Path of the file to check hash
    file_hash : str
        The expected hash string
    packet_size : int
        Bytes to read at a time for large files hashing.

    Returns
    -------
    validation : bool
    """
    hasher = hashlib.sha256()  # consider adding md5 option for legacy archives
    with open(filepath, "rb") as file:
        for packet in iter(lambda: file.read(packet_size), b""):
            hasher.update(packet)
    hashed = hasher.hexdigest()
    return str(hashed) == file_hash


def _url_retrieve(
    url: str, filename: str, url_data: Optional[Any] = None, progbar_disable=False
):
    """
    Download file given URL

    Parameters
    ----------
    url : str
        Url to retrieve.
    filename : str
        Local storage filename
    url_data : Optional[Any]
        Data argument passed to `urlopen`.
    progbar_disable : bool
        Disable progress bar. (default=False)
    """

    def packet_read(response, packet_size=8192):
        content_type = response.info().get("Content-Length")
        if content_type is not None:
            total_size = int(content_type.strip())
        else:  # pragma: no cover
            raise FileNotFoundError("Invalid content. Please check your URL and file")
            total_size = -1
        progbar = tqdm(total=total_size // 1024**1, disable=progbar_disable)
        while True:
            packet = response.read(packet_size)
            progbar.update(packet_size)
            if packet:
                yield packet
            else:
                progbar.update(total_size - progbar.n)
                progbar.close()
                break

    response = urlopen(url, url_data)
    with open(filename, "wb") as fd:
        for packet in packet_read(response):
            fd.write(packet)


def _extract_archive(
    file_path, output_path: str = ".", archive_format: str = "zip"
) -> bool:
    """
    Extracts an archive.
    Supports .tar (tar, tar.gz, tar.bz) or .zip format.

    If any error is encountered, the downloaded file is deleted.

    Parameters
    ----------
    file_path : str
        path to the archive file
    output_path : str
        path to extract the archive file
    archive_format : str
        Format of archived file. (default='zip')
        Options are ['tar','tar.gz','tar.bz','zip'] files.

    Returns
    -------
    Completion : bool
        return true if an archive extraction was completed,
    """
    assert archive_format in [
        "zip",
        "tar",
        "tar.gz",
        "tar.bz",
    ], "Provided archive format is not yet supported."
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"Archive file ({file_path}) does not exist.")

    if archive_format.startswith("tar"):
        open_fn = tarfile.open
        is_match_fn = tarfile.is_tarfile
    elif archive_format == "zip":
        open_fn = zipfile.ZipFile
        is_match_fn = zipfile.is_zipfile

    if is_match_fn(file_path):
        with open_fn(file_path, "r") as archive:
            try:
                # Extraction attempts
                archive.extractall(output_path)
            except (
                zipfile.BadZipFile,
                tarfile.TarError,
                RuntimeError,
                KeyboardInterrupt,
            ):  # pragma: no cover
                # On failure, remove any residue packets
                if os.path.exists(output_path):
                    if os.path.isfile(output_path):
                        os.remove(output_path)
                    else:
                        shutil.rmtree(output_path)
                raise RuntimeError(
                    "Failure/Interrupted during extracting archive-file."
                )
        return True
    return False



from miv.visualization.causality import *
from miv.visualization.connectivity import *
from miv.visualization.event import *
from miv.visualization.fft_domain import *
from miv.visualization.raw_signal import *
from miv.visualization.waveform import *

# TODO:
# Unit testing in visualization is simply performed such that the file or image is properly generated.
# It does not check the correctness of the visualization. The inspection is done manually.
# If anyone have better idea on how to do unittest on visualization, please help us out.



import os

import matplotlib.pyplot as plt
import numpy as np

from miv.statistics.spiketrain_statistics import interspike_intervals
from miv.typing import SpikestampsType


def burst(spiketrains: SpikestampsType, channel: float, min_isi: float, min_len: float):
    """
    Calculates parameters critical to characterize bursting phenomenon on a single channel
    Bursting is defined as the occurence of a specified number of spikes (usually >10), with a small interspike interval (usually < 100ms) [1]_, [2]_

    Parameters
    ----------
    spikes : SpikestampsType
       Single spike-stamps
    Channel : float
       Channel to analyze
    min_isi : float
       Minimum Interspike Interval (in seconds) to be considered as bursting [standard = 0.1]
    min_len : float
       Minimum number of simultaneous spikes to be considered as bursting [standard = 10]

    Returns
    -------
    start_time: float
       The time instances when a burst starts
    burst_duration: float
       The time duration of a particular burst
    burst_len: float
       Number of spikes in a particular burst
    burst_rate: float
       firing rates corresponding to particular bursts

    .. [1] Chiappalone, Michela, et al. "Burst detection algorithms for the analysis of spatio-temporal patterns
    in cortical networks of neurons." Neurocomputing 65 (2005): 653-662.
    .. [2] Eisenman, Lawrence N., et al. "Quantification of bursting and synchrony in cultured
    hippocampal neurons." Journal of neurophysiology 114.2 (2015): 1059-1071.
    """

    spike_interval = interspike_intervals(spiketrains[channel])
    assert spike_interval.all() > 0, "Inter Spike Interval cannot be zero"
    burst_spike = (spike_interval <= min_isi).astype(
        np.bool_
    )  # Only spikes within specified min ISI are 1 otherwise 0 and are stored

    # Try to find the start and end indices for burst interval
    delta = np.logical_xor(burst_spike[:-1], burst_spike[1:])
    interval = np.where(delta)[0]
    if len(interval) % 2:
        interval = np.append(interval, len(delta))
    interval += 1
    interval = interval.reshape([-1, 2])
    mask = np.diff(interval) >= min_len
    interval = interval[mask.ravel(), :]
    Q = np.array(interval)

    if np.sum(Q) == 0:
        start_time = 0
        end_time = 0
        burst_duration = 0
        burst_rate = 0
        burst_len = 0
    else:
        spike = np.array(spiketrains[channel])
        start_time = spike[Q[:, 0]]
        end_time = spike[Q[:, 1]]
        burst_len = Q[:, 1] - Q[:, 0] + 1
        burst_duration = end_time - start_time
        burst_rate = burst_len / (burst_duration)

    return start_time, burst_duration, burst_len, burst_rate



# from typing import TypeAlias
from typing import List, Union

import neo
import numpy as np
import numpy.typing as npt

import miv
import miv.core

SignalType = Union[
    np.ndarray,
    neo.core.AnalogSignal,  # npt.DTypeLike
]  # Shape should be [signal_length, n_channel]
TimestampsType = np.ndarray
SpikestampsType = Union[np.ndarray, neo.core.SpikeTrain, miv.core.Spikestamps]
SpiketrainType = np.ndarray  # non-sparse boolean



__all__ = ["generate_random_spiketrain"]

from typing import List

import numpy as np
from neo.core import SpikeTrain
from quantities import s

from miv.typing import SpikestampsType


def generate_random_spiketrain(
    duration: float,
    spikes_per_second: float,
    num_channels: int,
    random_spike_strength: bool = False,
    spike_strength: int = 3,
    random_noise_multiplier: bool = False,
    noise_multiplier: float = 10,
) -> List[SpikestampsType]:
    """Generate spiketrains with noise

    Parameters
    ----------
    duration : float
        Length of recording in seconds
    spikes_per_second : float
        Number of spikes per second
    num_channels : int
        Number of channels
    random_spike_strength : bool
        Generates random spike strength values for each channel
    spike_strength : int
        The number of times each spike will be repeated
    random_noise_multiplier : bool
        Generates random noise multipliers for each channel
    noise_multiplier : float
        The amount of noise spikes. This value is the ratio of noise spikes to real spikes.

    Returns
    -------
    artifical_spikes : List[SpikestampsType]

    """

    artificial_spikes = []

    for chan in range(num_channels):
        # Step one: generate true spikes
        if random_spike_strength:
            spike_strength = int(np.random.rand() * 10)
        channel_spikes = np.repeat(
            np.arange(start=0, stop=duration, step=1 / spikes_per_second),
            spike_strength,
        )
        # Step two: add noise
        if random_noise_multiplier:
            noise_multiplier = 10 ** (np.random.rand() * 2)
        noise_count = int(spikes_per_second * duration * noise_multiplier)
        channel_spikes = np.concatenate(
            (channel_spikes, np.random.rand(noise_count) * duration)
        )

        artificial_spikes.append(SpikeTrain(channel_spikes * s, duration))

    return artificial_spikes



__all__ = ["MEAGeometryProtocol"]

import typing
from typing import Any, Iterable, Protocol, Tuple

import matplotlib


class MEAGeometryProtocol(Protocol):
    def get_closest_node(self, x: float, y: float) -> int:
        """Given xy coordinate, return closest node idx"""
        ...

    def get_xy(self, idx: int) -> Tuple[float, float]:
        """Given node index, return xy coordinate"""
        ...

    def save(self, path: str) -> None:
        ...

    def load(self, path: str) -> None:
        ...

    def view(self) -> matplotlib.pyplot.Figure:
        """Simplified view of MEA orientation"""
        ...



__doc__ = """Protocol for Data objects."""
__all__ = ["DataProtocol"]

import typing
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Optional,
    Protocol,
    Set,
    Tuple,
    Union,
)

import os

from miv.typing import SignalType, TimestampsType


class DataProtocol(Protocol):
    """Behavior definition for a single experimental data handler."""

    def __init__(self, data_path: str):
        ...

    @property
    def analysis_path(self) -> None:
        ...

    def load(self, *args) -> Generator[SignalType, TimestampsType, int]:
        """Iterator to load data fragmentally. Use to load large file size data."""
        ...

    def check_path_validity(self) -> bool:
        """Check if necessary files exist in the directory."""
        ...



#! /usr/bin/env python

# This file is part of the SpikeCoding repository - MAVLab TU Delft
#
# MIT License
#
# Copyright (c) 2021 Julien Dupeyroux
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# @author Julien Dupeyroux

import csv

import DecodingSchemes as DS
import EncodingSchemes as ES
import FunctionGenerator as FG
import matplotlib.pyplot as plt
import numpy as np
from scipy import signal
from scipy.stats import norm

if __name__ == "__main__":
    dt = 0.01
    T_max = 4
    time = np.arange(0, T_max, dt)

    S = list()
    S.append(
        FG.sum_of_sine_waves(
            [2, -0.5, 0.75], [1.0, 3.0, 5.0], [0.0, 0.0, 0.0], 0.0, time
        )
    )
    S.append(FG.sum_of_sine_waves([-0.25], [1.0], [0.0], 0.05, time))
    S.append(FG.sum_of_gaussian_wave([1, 0.5], [0.2, 0.75], [0.1, 0.1], 0.1, time))

    # tbr_factors = [1.005, 1.005, 1.005]

    # sf_thresholds = [0.35, 0.05, 0.35]

    # mw_thresholds = [0.325, 0.015, 0.225]
    # mw_window = [3, 3, 3]

    # for i in range(len(S)):

    #     spikes_TBR, threshold = ES.temporal_contrast(S[i], tbr_factors[i])
    #     signal_TBR = 2*DS.temporal_contrast(spikes_TBR, threshold)

    #     spikes_SF, startpoint = ES.step_forward(S[i], sf_thresholds[i])
    #     signal_SF = DS.step_forward(spikes_SF, sf_thresholds[i], startpoint)

    #     spikes_MW, startpoint = ES.moving_window(S[i], mw_thresholds[i], mw_window[i])
    #     signal_MW = DS.moving_window(spikes_MW, mw_thresholds[i], startpoint)

    #     plt.subplot(3*len(S),3,(1+i*3*len(S),4+i*3*len(S)))
    #     plt.plot(time, S[i])
    #     plt.plot(time, signal_TBR)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)
    #     if i == 0:
    #         plt.title("Temporal Contrast Algorithm TBR")

    #     plt.subplot(3*len(S),3,7+i*3*len(S))
    #     plt.stem(time, spikes_TBR)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)

    #     plt.subplot(3*len(S),3,(2+i*3*len(S),5+i*3*len(S)))
    #     plt.plot(time, S[i])
    #     plt.plot(time, signal_SF)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)
    #     if i == 0:
    #         plt.title("Step Forward Algorithm SF")

    #     plt.subplot(3*len(S),3,8+i*3*len(S))
    #     plt.stem(time, spikes_SF)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)

    #     plt.subplot(3*len(S),3,(3+i*3*len(S),6+i*3*len(S)))
    #     plt.plot(time, S[i])
    #     plt.plot(time, signal_MW)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)
    #     if i == 0:
    #         plt.title("Moving Window Algorithm MW")

    #     plt.subplot(3*len(S),3,9+i*3*len(S))
    #     plt.stem(time, spikes_MW)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)

    # plt.show()

    # hsa_window = [12, 15, 12]
    # hsa_fir = list()
    # hsa_fir.append(signal.triang(hsa_window[0]))
    # hsa_fir.append(norm.pdf(np.linspace(1, hsa_window[1], hsa_window[1]), 0, 5))
    # hsa_fir.append(signal.triang(hsa_window[2]))

    # hsa_m_thresholds = [0.85, 0.05, 0.5]

    # bsa_window = [9, 10, 8]
    # bsa_fir = list()
    # bsa_fir.append(signal.triang(bsa_window[0]))
    # bsa_fir.append(norm.pdf(np.linspace(1, bsa_window[1], bsa_window[1]), 1.5, 3.5))
    # bsa_fir.append(signal.triang(bsa_window[2]))

    # bsa_thresholds = [1.175, 1.05, 1.2]

    # for i in range(len(S)):

    #     spikes_HSA, shift = ES.hough_spike(S[i], hsa_fir[i])
    #     signal_HSA = DS.ben_spike(spikes_HSA, hsa_fir[i], shift)

    #     spikes_HSAm, shift = ES.modified_hough_spike(S[i], hsa_fir[i], hsa_m_thresholds[i])
    #     signal_HSAm = DS.ben_spike(spikes_HSAm, hsa_fir[i], shift)

    #     spikes_BSA, shift = ES.ben_spike(S[i], bsa_fir[i], bsa_thresholds[i])
    #     signal_BSA = DS.ben_spike(spikes_BSA, bsa_fir[i], shift)

    #     plt.subplot(3*len(S),3,(1+i*3*len(S),4+i*3*len(S)))
    #     plt.plot(time, S[i])
    #     plt.plot(time, signal_HSA)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)
    #     if i == 0:
    #         plt.title("Hough Spike Algorithm HSA")

    #     plt.subplot(3*len(S),3,7+i*3*len(S))
    #     plt.stem(time, spikes_HSA)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)

    #     plt.subplot(3*len(S),3,(2+i*3*len(S),5+i*3*len(S)))
    #     plt.plot(time, S[i])
    #     plt.plot(time, signal_HSAm)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)
    #     if i == 0:
    #         plt.title("Threshold Hough Spike Algorithm T-HSA")

    #     plt.subplot(3*len(S),3,8+i*3*len(S))
    #     plt.stem(time, spikes_HSAm)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)

    #     plt.subplot(3*len(S),3,(3+i*3*len(S),6+i*3*len(S)))
    #     plt.plot(time, S[i])
    #     plt.plot(time, signal_BSA)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)
    #     if i == 0:
    #         plt.title("Ben Spike Algorithm BSA")

    #     plt.subplot(3*len(S),3,9+i*3*len(S))
    #     plt.stem(time, spikes_BSA)
    #     plt.gca().axes.get_xaxis().set_visible(False)
    #     plt.gca().axes.get_yaxis().set_visible(False)

    # plt.show()

    number_of_neurons_grf = 15

    nb_neurons_BOHTE = 15
    nb_timesteps_BOHTE = 10
    beta_BOHTE = 1.5

    for i in range(len(S)):

        [min_input, max_input] = [min(S[i]), max(S[i])]
        spikes_GRF = ES.grf_spike(S[i], number_of_neurons_grf, min_input, max_input)
        signal_GRF = DS.grf_spike(spikes_GRF, min_input, max_input)

        spikes_BOHTE = ES.grf_spike_with_internal_timesteps(
            S[i], min_input, max_input, nb_neurons_BOHTE, nb_timesteps_BOHTE, beta_BOHTE
        )
        signal_BOHTE = DS.grf_spike_with_internal_timesteps(
            spikes_BOHTE, nb_timesteps_BOHTE, min_input, max_input
        )
        # shape = spikes_BOHTE.shape
        # new_spikes_BOHTE = spikes_BOHTE.reshape((int(shape[0]/nb_timesteps_BOHTE), nb_timesteps_BOHTE, shape[1]))

        plt.subplot(3 * len(S), 3, (1 + i * 3 * len(S), 4 + i * 3 * len(S)))
        plt.plot(time, S[i])
        plt.plot(time, signal_GRF)
        plt.gca().axes.get_xaxis().set_visible(False)
        plt.gca().axes.get_yaxis().set_visible(False)
        if i == 0:
            plt.title("Gaussian Receptive Fields GRF")

        plt.subplot(3 * len(S), 3, 7 + i * 3 * len(S))
        for k in range(len(S[i])):
            plt.plot(
                [time[k], time[k]],
                [
                    np.argmax(spikes_GRF[k, :]) - 0.35,
                    np.argmax(spikes_GRF[k, :]) + 0.35,
                ],
            )
        plt.gca().axes.get_xaxis().set_visible(False)
        plt.gca().axes.get_yaxis().set_visible(False)

        plt.subplot(3 * len(S), 3, (3 + i * 3 * len(S), 6 + i * 3 * len(S)))
        plt.plot(time, S[i])
        plt.plot(time, signal_BOHTE)
        plt.gca().axes.get_xaxis().set_visible(False)
        plt.gca().axes.get_yaxis().set_visible(False)
        if i == 0:
            plt.title("Bohté et al. (2002) GRF")

        plt.subplot(3 * len(S), 3, 9 + i * 3 * len(S))
        # TODO: add plot
        plt.gca().axes.get_xaxis().set_visible(False)
        plt.gca().axes.get_yaxis().set_visible(False)

    plt.show()



__all__ = ["BensSpikerAlgorithm"]
from typing import Tuple

import numpy as np
from scipy.signal import firwin
from sklearn.preprocessing import normalize

from miv.typing import SignalType, TimestampsType


class BensSpikerAlgorithm:
    """
    Bens Spiker Algorithm (BSA) Implementation.
    Implementation based on [1]_ [2]_ [3]_

    .. [1] Petro et al. (2020)
    .. [2] Sengupta et al. (2017)
    .. [3] Schrauwen, B., et al., BSA, a fast and accurate spike train encoding scheme. IJCNN (2003)

    Examples
    --------

        >>> bsa_spiker: SpikerAlgorithmProtocol  = BensSpikerAlgorithm()
        >>> bsa_spiker(data)
        [[[0 1 1 ... 1 0 0]
          [0 1 1 ... 0 0 1]
          ...
          [0 0 0 ... 0 1 0]]
         [[0 0 1 ... 1 0 1]
          [1 0 1 ... 0 1 1]
          ...
          [0 0 1 ... 1 0 0]]]
    """

    def __init__(
        self,
        sampling_rate: float,
        threshold: float = 1.0,
        fir_filter_length: int = 2,
        fir_cutoff: float = 0.8,
        normalize: bool = False,
    ):
        """

        Parameters
        ----------
        sampling_rate : float
            Sampling rate in float. Used to compute timestamps to return.
            Note, this might not be the sampling rate of the signal, if signal passed
            through auditory model.
        threshold : float or Iterable[float]
            Threshold level for spike. It can also be a list of threshold.
            If list of threshold is given, the length of the list must be
            the number of features/channels. (default=1.0)
        fir_filter_length : int
            Filter_length for finite impulse response (FIR) filter (default=2)
        fir_cutoff : float
            Cutoff for finite impulse response (FIR) filter (default=0.8)
        normalize : bool
            Normalize data before comparing against threshold. (default=False)
            If true, input data is normalized in range of [0,1].
            IF false, input data is only shifted to have range of [0,~]
        """
        super().__init__()
        self.threshold = threshold
        self.fir_filter_length = fir_filter_length
        self.fir_cutoff = fir_cutoff
        self.data_normalize = normalize
        self.sampling_rate = sampling_rate

    def __call__(self, data: np.ndarray, time_offset: float = 0) -> np.ndarray:
        """
        Returns spiketrain after applying BSA to the data.

        Parameters
        ----------
        data : np.ndarray
            shape (length, n_channels)
        time_offset : float
            time offset for returning timestamps (default=0)

        Returns
        -------
        Spiketrains : SpiketrainType
        timestamps: TimestampsType

        """

        # Data prep
        assert (
            len(data.shape) == 2
        ), "The shape of the data must be (length, n_channels)."
        data_length, num_channels = data.shape

        # Set thresholds
        if isinstance(self.threshold, list):
            assert num_channels == len(
                self.threshold
            ), "If threshold is given as a list, \
                    the length of the threshold must be same as the number of channels in data."
        elif isinstance(self.threshold, float) or isinstance(self.threshold, int):
            thresholds = np.ones(num_channels) * self.threshold
        else:
            raise TypeError(
                f"The threshold type {type(self.threshold)} is not supported."
            )

        # FIR
        filter_values, filter_length = self._finite_impulse_response(data)

        # Data standardization
        if self.data_normalize:
            _data = normalize(data, axis=1)
        else:
            _data = data.copy()
            _data -= data.min(axis=0, keepdims=True)

        spikes = np.zeros_like(data, dtype=np.bool_)

        # BSA Algorithm
        for ch in range(0, num_channels):  # TODO: Multithreaded
            for i in range(0, data_length - filter_length + 1):
                error1 = 0
                error2 = 0
                for j in range(0, filter_length):
                    error1 += abs(_data[i + j][ch] - filter_values[j])
                    error2 += abs(_data[i + j - 1][ch])

                if error1 <= error2 * thresholds[ch]:
                    spikes[i][ch] = True
                    for j in range(0, filter_length):
                        if i + j + 1 < data_length:
                            _data[i + j + 1][ch] -= filter_values[j]

        # Timestamps
        timestamps = np.arange(data_length) * (1 / self.sampling_rate) + time_offset

        return spikes, timestamps

    def _finite_impulse_response(self, data: np.ndarray) -> Tuple[np.ndarray, int]:
        """
        FIR filter

        Returns
        -------
        filtered_value
        filtered_length
        """
        filter_values = firwin(self.fir_filter_length, cutoff=self.fir_cutoff)
        return filter_values, len(filter_values)



__doc__ = """Sample data for criticality analysis."""
__all__ = ["load_data"]

import gzip
import os

import numpy as np

from miv.datasets.utils import get_file


def load_data():  # pragma: no cover
    """
    Loads the sample for criticality analysis

    Total size: 5.1 MB (compressed)

    Returns
    -------
    datapath: str

    Examples
    --------
        >>> from miv import datasets
        >>> data = datasets.criticality.load_data()
    """

    subdir = "criticality"
    base_url = "https://uofi.box.com/shared/static/mxj6mfjj0fdexod4nqoqovdeg7t8dbbq.mat"
    file = "asdf.mat"
    file_hash = "c8ff82f4700eed2023d37cb7643d24b37350e521469148406042ef206c3aac34"

    path = get_file(
        file_url=base_url,
        directory=subdir,
        fname=file,
        file_hash=file_hash,
        archive_format=None,
    )
    return path



__all__ = [
    "SpikeDetectionProtocol",
    "SpikeFeatureExtractionProtocol",
    "UnsupervisedFeatureClusteringProtocol",
]

from typing import Any, Iterable, Protocol, Union

import neo.core
import numpy as np

from miv.typing import SignalType, SpikestampsType, TimestampsType


class SpikeDetectionProtocol(Protocol):
    def __call__(
        self, signal: SignalType, timestamps: TimestampsType, sampling_rate: float
    ) -> SpikestampsType:
        ...

    def __repr__(self) -> str:
        ...


# TODO: Behavior is clear, but not sure what the name should be
# class SpikeSortingProtocol(Protocol):
#    def __call__(
#        self, signals: Iterable[SignalType]
#    ) -> np.ndarray:
#        ...
#
#    def __repr__(self) -> str:
#        ...


class SpikeFeatureExtractionProtocol(Protocol):
    """ex) wavelet transform, PCA, etc."""

    def __repr__(self) -> str:
        ...


class UnsupervisedFeatureClusteringProtocol(Protocol):
    def __repr__(self) -> str:
        ...

    def fit(self, X: np.ndarray):
        ...

    def predict(self, X: np.ndarray) -> np.ndarray:
        ...



__all__ = ["plot_connectivity", "plot_connectivity_interactive"]

import os

import graphviz
import matplotlib.pyplot as plt
import numpy as np
from pyvis.network import Network

from miv.typing import SpikestampsType


def plot_connectivity(
    mea_map: np.ndarray, connectivity_matrix: np.ndarray, directionality: bool
):
    """
    Plots the provided connectivity matrix with graphviz using the exact location of electrode.

    Parameters
    ----------
    mea_map : np.ndarray
        array containing spatial location of electrodes
    connectivity_matrix: np.ndarray
        array containing the connectivity parameters for each pair of electrodes
    directionality: bool
        if set true can be used to plot directional connectivity plots

    Returns
    -------
    g
        graphviz graph object
    """

    connec_x, connec_y = np.shape(connectivity_matrix)
    assert connec_x == connec_y, "Connectivity matrix should be a square matrix"
    num_elec = np.count_nonzero(mea_map)
    assert num_elec > 0, "Number of electrodes should be greater than 0"
    elec_mat = np.linspace(1, num_elec, num_elec).astype(int)
    assert (
        connec_x == num_elec
    ), "Connectivity matrix should have same dimensions as the number of electrodes"

    if directionality:
        g = graphviz.Digraph("G", filename="connectivity", engine="neato")
    else:
        g = graphviz.Graph("G", filename="connectivity", engine="neato")

    for i in elec_mat:
        for j in elec_mat:
            if i == j:
                continue
            if not directionality and i >= j:
                continue
            if np.isclose(connectivity_matrix[int(i - 1), int(j - 1)], 0):
                continue
            # Register
            ki = int(np.where(mea_map == i)[0][0])
            ji = int(np.where(mea_map == i)[1][0])
            kj = int(np.where(mea_map == j)[0][0])
            jj = int(np.where(mea_map == j)[1][0])

            x1 = str(ki + 0.0)
            y1 = str(ji) + "!"
            posstr1 = x1 + "," + y1

            x2 = str(kj + 0.0)
            y2 = str(jj) + "!"
            posstr2 = x2 + "," + y2
            g.edge(str(i), str(j), color="red")

            g.node(
                str(i),
                pos=posstr1,
                fillcolor="deepskyblue2",
                color="black",
                style="filled,solid",
                shape="circle",
                fontcolor="black",
                fontsize="15",
                fontname="Times:Roman bold",
            )

            g.node(
                str(j),
                pos=posstr2,
                fillcolor="deepskyblue2",
                color="black",
                style="filled,solid",
                shape="circle",
                fontcolor="black",
                fontsize="15",
                fontname="Times:Roman bold",
            )
    return g


def plot_connectivity_interactive(
    mea_map: np.ndarray,
    connectivity_matrix: np.ndarray,
    directionality: bool,
):
    """
    Plots the provided connectivity matrix with pyvis using the exact location of electrode.

    Parameters
    ----------
    mea_map : np.ndarray
        array containing spatial location of electrodes
    connectivity_matrix: np.ndarray
        array containing the connectivity parameters for each pair of electrodes
    directionality: bool
        if set true can be used to plot directional connectivity plots

    Returns
    -------
    net
        pyvis network object
    """
    connec_x, connec_y = np.shape(connectivity_matrix)
    assert connec_x == connec_y, "Connectivity matrix should be a square matrix"
    num_elec = np.count_nonzero(mea_map)
    assert num_elec > 0, "Number of electrodes should be greater than 0"
    elec_mat = np.linspace(1, num_elec, num_elec).astype(int)
    assert (
        connec_x == num_elec
    ), "Connectivity matrix should have same dimensions as the number of electrodes"

    net = Network(
        height="500px",
        width="100%",
        bgcolor="#222222",
        directed=directionality,
        font_color="white",
        notebook="false",
    )
    net.repulsion()

    for i in elec_mat:
        for j in elec_mat:
            if i == j:
                continue
            if np.isclose(connectivity_matrix[int(i - 1), int(j - 1)], 0.0) or i >= j:
                continue
            ki = int(np.where(mea_map == i)[0][0]) * 100
            ji = int(np.where(mea_map == i)[1][0]) * 100
            kj = int(np.where(mea_map == j)[0][0]) * 100
            jj = int(np.where(mea_map == j)[1][0]) * 100

            x1 = str(ki + 0.0)
            y1 = str(ji)

            x2 = str(kj + 0.0)
            y2 = str(jj)

            net.add_node(
                str(int(i)), x=x1, y=y1, shape="dot", color="#039AFB"
            )  # size = size1)
            net.add_node(str(int(j)), x=x2, y=y2, shape="dot", color="#039AFB!")
            net.add_edge(str(int(i)), str(int(j)), width="1")

    for n in net.nodes:
        n.update({"physics": False})

    return net



__doc__ = """
Module for extracting each spike waveform and visualize.
"""
__all__ = ["extract_waveforms", "plot_waveforms"]

from typing import Any, Dict, Optional, Tuple, Union

import os

import matplotlib
import matplotlib.pyplot as plt
import neo
import numpy as np
import quantities as pq
from scipy.signal import lfilter, savgol_filter
from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import StandardScaler

from miv.typing import SignalType, SpikestampsType

# TODO: Modularize the entire process.


def extract_waveforms(
    signal: SignalType,
    spikestamps: SpikestampsType,
    channel: Optional[int],
    sampling_rate: float,
    pre: pq.Quantity = 0.001 * pq.s,
    post: pq.Quantity = 0.002 * pq.s,
) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    Extract spike waveforms as signal cutouts around each spike index as a spikes x samples numpy array

    Parameters
    ----------
    signal : SignalType
        The signal as a 2-dimensional numpy array (length, num_channel)
    spikestamps : SpikestampsType
        The sample index of all spikes as a 1-dim numpy array
    channel : Optional[int]
        Interested channel. If None, assume signal and spikestamps are single channel.
    sampling_rate : float
        The sampling frequency in Hz
    pre : pq.Quantity
        The duration of the cutout before the spike in seconds. (default=0.001 s)
    post : pq.Quantity
        The duration of the cutout after the spike in seconds. (default=0.002 s)

    Returns
    -------
    Stack of spike cutout: np.ndarray
        Return stacks of spike cutout; shape(n_spikes, width).

    """
    if channel is not None:
        signal = signal[:, channel]
        spikestamps = spikestamps[channel]

    cutouts = []
    pre_idx = int(pre * sampling_rate)
    post_idx = int(post * sampling_rate)

    assert (
        pre_idx + post_idx > 0
    ), "Set larger pre/post duration. pre+post duration must be more than 1/sampling_rate."

    # Padding signal
    signal = np.pad(signal, ((pre_idx, post_idx),), constant_values=0)
    for time in spikestamps:
        index = int(round(time * sampling_rate))
        if index >= signal.shape[0] or index + post_idx + pre_idx >= signal.shape[0]:
            raise IndexError(
                "The width of the spike exceeded the signal. "
                "Either timestamp exceeded the maximum time recorded, or "
                "post duration is too large."
            )
        # if index - pre_idx >= 0 and index + post_idx <= signal.shape[0]:
        #    cutout = signal[(index - pre_idx) : (index + post_idx)]
        #    cutouts.append(cutout)
        cutout = signal[index : (index + post_idx + pre_idx)]
        cutouts.append(cutout)

    return np.stack(cutouts)


def plot_waveforms(
    cutouts: np.ndarray,
    sampling_rate: float,
    pre: float = 0.001,
    post: float = 0.002,
    n_spikes: Optional[int] = 100,
    color: str = "k",  # TODO: change typing to matplotlib color
    ax: Optional[matplotlib.axes.Axes] = None,
    return_time: bool = False,
    plot_kwargs: Dict[Any, Any] = None,
) -> plt.Figure:
    """
    Plot an overlay of spike cutouts

    Parameters
    ----------
    cutouts : np.ndarray
        A spikes x samples array of cutouts
    sampling_rate : float
        The sampling frequency in Hz
    pre : float
        The duration of the cutout before the spike in seconds
    post : float
        The duration of the cutout after the spike in seconds
    n_spikes : Optional[int]
        The number of cutouts to plot. None to plot all. (Default: 100)
    color : str
        The line color as a pyplot line/marker style. (Default: 'k'=black)
    ax : Optional[matplotlib.axes.Axes]
        Use provided axes to plot. If none, create new figure.
    plot_kwargs : Dict[Any, Any]
        Addtional keyword-arguments for matplotlib.pyplot.plot.
    """
    if n_spikes is None:
        n_spikes = cutouts.shape[0]
    n_spikes = min(n_spikes, cutouts.shape[0])

    if not plot_kwargs:
        plot_kwargs = {"alpha": 0.3, "linewidth": 1}

    # TODO: Need to match unit
    time = np.arange(-pre * 1000, post * 1000, 1e3 / sampling_rate)

    fig = plt.gcf()
    if ax is None:
        ax = plt.gca()
    for i in range(n_spikes):
        ax.plot(
            time,
            cutouts[
                i,
            ],
            color,
            **plot_kwargs,
        )
    ax.set_xlabel("Time (ms)")
    ax.set_ylabel("Voltage (uV)")
    ax.set_title("Cutouts")
    if return_time:
        return fig, time
    return fig



__doc__ = """

Spike Detection
###############

Code Example::

    detection = ThresholdCutoff(cutoff=3.5)
    spiketrains = detection(signal, timestamps, sampling_rate)


.. currentmodule:: miv.signal.spike

.. autosummary::
   :nosignatures:
   :toctree: _toctree/DetectionAPI

   SpikeDetectionProtocol
   ThresholdCutoff

"""
__all__ = ["ThresholdCutoff"]

from typing import Iterable, List, Union

from dataclasses import dataclass

import neo
import numpy as np
import quantities as pq
from tqdm import tqdm

from miv.core import Spikestamps
from miv.typing import SignalType, SpikestampsType, TimestampsType


@dataclass
class ThresholdCutoff:
    """ThresholdCutoff
    Spike sorting step by step guide is well documented `here <http://www.scholarpedia.org/article/Spike_sorting>`_.

        Attributes
        ----------
        dead_time : float
            (default=0.003)
        search_range : float
            (default=0.002)
        cutoff : Union[float, np.ndarray]
            (default=5.0)
        use_mad : bool
            (default=False)
        tag : str
    """

    dead_time: float = 0.003
    search_range: float = 0.002
    cutoff: float = 5.0
    use_mad: bool = True
    tag: str = "Threshold Cutoff Spike Detection"

    def __call__(
        self,
        signal: SignalType,
        timestamps: TimestampsType,
        sampling_rate: float,
        units: Union[str, pq.UnitTime] = "sec",
        progress_bar: bool = True,
        return_neotype: bool = True,
    ) -> SpikestampsType:
        """Execute threshold-cutoff method and return spike stamps

        Parameters
        ----------
        signal : SignalType
            signal
        timestamps : TimestampsType
            timestamps
        sampling_rate : float
            sampling_rate
        units : Union[str, pq.UnitTime]
            (default='sec')
        progress_bar : bool
            Toggle progress bar (default=True)
        return_neotype : bool
            If true, return spiketrains in neo.Spiketrains (default=True)
            If false, return list of numpy-type spiketrains.

        Returns
        -------
        spiketrain_list : List[SpikestampsType]

        """
        # Spike detection for each channel
        spiketrain_list = []
        num_channels = signal.shape[1]  # type: ignore
        for channel in tqdm(range(num_channels), disable=not progress_bar):
            array = signal[:, channel]  # type: ignore

            # Spike Detection: get spikestamp
            spike_threshold = self.compute_spike_threshold(
                array, cutoff=self.cutoff, use_mad=self.use_mad
            )
            crossings = self.detect_threshold_crossings(
                array, sampling_rate, spike_threshold, self.dead_time
            )
            spikes = self.align_to_minimum(
                array, sampling_rate, crossings, self.search_range
            )
            spikestamp = spikes / sampling_rate + timestamps.min()
            # Convert spikestamp to neo.SpikeTrain (for plotting)
            if return_neotype:
                spiketrain = neo.SpikeTrain(
                    spikestamp,
                    units=units,
                    t_stop=timestamps.max(),
                    t_start=timestamps.min(),
                )
                spiketrain_list.append(spiketrain)
            else:
                spiketrain_list.append(spikestamp.astype(np.float_))
        return Spikestamps(spiketrain_list)

    def compute_spike_threshold(
        self, signal: SignalType, cutoff: float = 5.0, use_mad: bool = True
    ) -> float:  # TODO: make this function compatible to array of cutoffs (for each channel)
        """
        Returns the threshold for the spike detection given an array of signal.

        Spike sorting step by step, `step 2 <http://www.scholarpedia.org/article/Spike_sorting>`_.

        Parameters
        ----------
        signal : np.array
            The signal as a 1-dimensional numpy array
        cutoff : float
            The spike-cutoff multiplier. (default=5.0)
        use_mad : bool
            Noise estimation method. If set to false, use standard deviation for estimation. (default=True)
        """
        if use_mad:
            noise_mid = np.median(np.absolute(signal)) / 0.6745
        else:
            noise_mid = np.std(signal)
        spike_threshold = -cutoff * noise_mid
        return spike_threshold

    def detect_threshold_crossings(
        self, signal: SignalType, fs: float, threshold: float, dead_time: float
    ):
        """
        Detect threshold crossings in a signal with dead time and return them as an array

        The signal transitions from a sample above the threshold to a sample below the threshold for a detection and
        the last detection has to be more than dead_time apart from the current one.

        Parameters
        ----------
        signal : SignalType
            The signal as a 1-dimensional numpy array
        fs : float
            The sampling frequency in Hz
        threshold : float
            The threshold for the signal
        dead_time : float
            The dead time in seconds.
        """
        dead_time_idx = dead_time * fs
        threshold_crossings = np.diff((signal <= threshold).astype(int) > 0).nonzero()[
            0
        ]
        distance_sufficient = np.insert(
            np.diff(threshold_crossings) >= dead_time_idx, 0, True
        )
        while not np.all(distance_sufficient):
            # repeatedly remove all threshold crossings that violate the dead_time
            threshold_crossings = threshold_crossings[distance_sufficient]
            distance_sufficient = np.insert(
                np.diff(threshold_crossings) >= dead_time_idx, 0, True
            )
        return threshold_crossings

    def get_next_minimum(self, signal, index, max_samples_to_search):
        """
        Returns the index of the next minimum in the signal after an index

        :param signal: The signal as a 1-dimensional numpy array
        :param index: The scalar index
        :param max_samples_to_search: The number of samples to search for a minimum after the index
        """
        search_end_idx = min(index + max_samples_to_search, signal.shape[0])
        min_idx = np.argmin(signal[index:search_end_idx])
        return index + min_idx

    def align_to_minimum(self, signal, fs, threshold_crossings, search_range):
        """
        Returns the index of the next negative spike peak for all threshold crossings

        :param signal: The signal as a 1-dimensional numpy array
        :param fs: The sampling frequency in Hz
        :param threshold_crossings: The array of indices where the signal crossed the detection threshold
        :param search_range: The maximum duration in seconds to search for the minimum after each crossing
        """
        search_end = int(search_range * fs)
        aligned_spikes = [
            self.get_next_minimum(signal, t, search_end) for t in threshold_crossings
        ]
        return np.array(aligned_spikes)



__doc__ = """

Typical spike-sorting procedure can be described in three steps: (1) spike detection, (2) feature decomposition, and (3) clustering.
We provide separate module to perform spike-detection; see :ref:`here <api/signal:Spike Detection>`.

We provide `SpikeSorting` module that composes *feature-decomposition* method and *unsupervised-clustering* method.
A commonly used feature-decomposition method includes PCA or wavelet decomposition.
For clustering method, one implemented few commonly appearing methods from the literatures (listed below).
Additionally, one can use out-of-the-box clustering modules from `sklearn`.

.. note:: Depending on the method of clustering, there might be an additional step to find optimum number of cluster.

.. currentmodule:: miv.signal.spike

.. autoclass:: SpikeSorting
   :members:

Available Feature Extractor
###########################

.. autosummary::
   :toctree: _toctree/SpikeSortingAPI

   SpikeFeatureExtractionProtocol
   WaveletDecomposition
   PCADecomposition

Unsupervised Clustering
#######################

.. autosummary::
   :toctree: _toctree/SpikeSortingAPI

   UnsupervisedFeatureClusteringProtocol
   SuperParamagneticClustering

Other external tools
--------------------

Following external modules can also be used for the spike sorting.

Sklearn Clustering
~~~~~~~~~~~~~~~~~~
.. autosummary::

   sklearn.cluster.MeanShift
   sklearn.cluster.KMeans

"""
__all__ = [
    "SpikeSorting",
    "WaveletDecomposition",
    "PCADecomposition",
    "SuperParamagneticClustering",
]

from typing import Any, Optional, Union

from dataclasses import dataclass

import matplotlib.pyplot as plt
import neo
import numpy as np
import pywt
import quantities as pq
import scipy
import scipy.special
from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture
from sklearn.preprocessing import StandardScaler

from miv.signal.spike.protocol import (
    SpikeFeatureExtractionProtocol,
    UnsupervisedFeatureClusteringProtocol,
)
from miv.typing import SignalType, SpikestampsType, TimestampsType


class SpikeSorting:
    """
    Spike sorting module.
    User can specify the method for feature extraction (e.g. WaveletDecomposition, PCADecomposition, etc)
    and the method for clustering (e.g. MeanShift, KMeans, SuperParamagneticClustering, etc).


    Examples
    --------
    >>> spike_sorting = SpikeSorting(
    ...    feature_extractor=WaveletDecomposition(),
    ...    clustering_method=sklearn.cluster.MeanShift()
    ... )
    >>> label, index = spike_sorting(cutouts, return_index=True)


    Parameters
    ----------
    feature_extractor : SpikeFeatureExtractionProtocol
    clustering_method : UnsupervisedFeatureClusteringProtocol

    """

    def __init__(
        self,
        feature_extractor: SpikeFeatureExtractionProtocol,
        clustering_method: UnsupervisedFeatureClusteringProtocol,
    ):
        self.featrue_extractor = feature_extractor
        self.clustering_method = clustering_method

    def __call__(self, cutouts: np.ndarray, n_group: int = 3):
        assert n_group >= 2, "n_group must be larger than 1"


# UnsupervisedFeatureClusteringProtocol
class SuperParamagneticClustering:  # pragma : no cover
    """Super-Paramagnetic Clustering (SPC)

    The implementation is heavily inspired from [1]_ and [2]_.


    .. [1] Quiroga RQ, Nadasdy Z, Ben-Shaul Y. Unsupervised spike detection and sorting with wavelets and superparamagnetic clustering. Neural Comput. 2004 Aug;16(8):1661-87. doi: 10.1162/089976604774201631. PMID: 15228749.
    .. [2] Fernando J. Chaure, Hernan G. Rey, and Rodrigo Quian Quiroga. A novel and fully automatic spike-sorting implementation with variable number of features. Journal of Neurophysiology 2018 120:4, 1859-1871. https://doi.org/10.1152/jn.00339.2018

    """

    pass


class PCADecomposition:
    """PCA Decomposition

    Other studies that use PCA decomposition: [1]_, [2]_

    .. [1] G. Hilgen, M. Sorbaro, S. Pirmoradian, J.-O. Muthmann, I. Kepiro, S. Ullo, C. Juarez Ramirez, A. Puente Encinas, A. Maccione, L. Berdondini, V. Murino, D. Sona, F. Cella Zanacchi, E. Sernagor, M.H. Hennig (2016). Unsupervised spike sorting for large scale, high density multielectrode arrays. Cell Reports 18, 2521–2532. bioRxiv: http://dx.doi.org/10.1101/048645.
    .. [2] Yger P, Spampinato GL, Esposito E, Lefebvre B, Deny S, Gardella C, Stimberg M, Jetter F, Zeck G, Picaud S, Duebel J, Marre O. A spike sorting toolbox for up to thousands of electrodes validated with ground truth recordings in vitro and in vivo. Elife. 2018 Mar 20;7:e34518. doi: 10.7554/eLife.34518. PMID: 29557782; PMCID: PMC5897014.

    """

    def __init__(self):
        pass

    def project(self, n_components, cutouts):
        scaler = StandardScaler()
        scaled_cutouts = scaler.fit_transform(cutouts)

        pca = PCA()
        pca.fit(scaled_cutouts)
        # print(pca.explained_variance_ratio_)

        pca.n_components = 2
        transformed = pca.fit_transform(scaled_cutouts)

        # Clustering
        gmm = GaussianMixture(n_components=n_components, n_init=10)
        labels = gmm.fit_predict(transformed)
        return labels, transformed

        """
        tmp_list = []
        for i in range(n_components):
            idx = labels == i
            tmp_list.append(timestamps[idx])
            spikestamps_clustered.append(tmp_list)

        _ = plt.figure(figsize=(8, 8))
        for i in range(n_components):
            idx = labels == i
            _ = plt.plot(transformed[idx, 0], transformed[idx, 1], ".")
            _ = plt.title("Cluster assignments by a GMM")
            _ = plt.xlabel("Principal Component 1")
            _ = plt.ylabel("Principal Component 2")
            _ = plt.legend([0, 1, 2])
            _ = plt.axis("tight")

        _ = plt.figure(figsize=(8, 8))
        for i in range(n_components):
            idx = labels == i
            color = plt.rcParams["axes.prop_cycle"].by_key()["color"][i]
            plot_waveforms(
                cutouts[idx, :],
                rate,
                n_spikes=100,
                color=color,
            )
        # custom legend
        custom_lines = [
            plt.Line2D(
                [0],
                [0],
                color=plt.rcParams["axes.prop_cycle"].by_key()["color"][i],
                lw=4,
            )
            for i in range(n_components)
        ]
        plt.legend(custom_lines, [f"component {i}" for i in range(n_components)])
        """


class WaveletDecomposition:  # TODO
    """
    Wavelet Decomposition for spike sorting.
    The implementation is heavily inspired from [1]_ and [2]_;
    their MatLab implementation (wave_clus) can be found `here <https://github.com/csn-le/wave_clus>`_.

    The default setting uses four-level multiresolution decomposition with Haar wavelets.
    To learn about possible choice of wavelet, check `PyWavelets module <https://pywavelets.readthedocs.io/en/latest/#>`_.

    Other studies that use wavelet decomposition: [3]_

    .. [1] Letelier JC, Weber PP. Spike sorting based on discrete wavelet transform coefficients. J Neurosci Methods. 2000 Sep 15;101(2):93-106. doi: 10.1016/s0165-0270(00)00250-8. PMID: 10996370.
    .. [2] Quiroga RQ, Nadasdy Z, Ben-Shaul Y. Unsupervised spike detection and sorting with wavelets and superparamagnetic clustering. Neural Comput. 2004 Aug;16(8):1661-87. doi: 10.1162/089976604774201631. PMID: 15228749.
    .. [3] Nenadic, Z., and Burdick, J. W. (2005). Spike detection using the continuous wavelet transform. IEEE Trans. BioMed. Eng. 52, 74–87. doi: 10.1109/TBME.2004.839800

    """

    def __init__(self):
        pass

    def project(self, n_features):
        ## Wavelet Decomposition
        number_of_spikes = 400
        data_length = 100
        cutouts = np.empty([number_of_spikes, data_length])
        # spikes_l = cutouts[0]
        coeffs = pywt.wavedec(cutouts, "haar", level=4)
        features = np.concatenate(coeffs, axis=1)

        def test_ks(x):
            # Calculates CDF

            # xCDF = []
            yCDF = []
            x = x[~np.isnan(x)]
            n = x.shape[0]
            x.sort()

            # Get cumulative sums
            yCDF = (np.arange(n) + 1) / n

            # Remove duplicates; only need final one with total count
            notdup = np.concatenate([np.diff(x), [1]]) > 0
            x_expcdf = x[notdup]
            y_expcdf = np.concatenate([[0], yCDF[notdup]])

            # The theoretical CDF (theocdf) is assumed to ben ormal
            # with unknown mean and sigma
            zScore = (x_expcdf - x.mean()) / x.std()
            # theocdf = normcdf(zScore, 0, 1)

            mu = 0
            sigma = 1
            theocdf = 0.5 * scipy.special.erfc(-(zScore - mu) / (np.sqrt(2) * sigma))

            # Compute the Maximum distance: max|S(x) - theocdf(x)|.

            delta1 = (
                y_expcdf[:-1] - theocdf
            )  # Vertical difference at jumps approaching from the LEFT.
            delta2 = (
                y_expcdf[1:] - theocdf
            )  # Vertical difference at jumps approaching from the RIGHT.
            deltacdf = np.abs(np.concatenate([delta1, delta2]))

            KSmax = deltacdf.max()
            return KSmax

        ks = []
        for idx, feature in enumerate(np.moveaxis(features, 1, 0)):
            std_feature = np.std(feature)
            mean_feature = np.mean(feature)
            thr_dist = std_feature * 3
            thr_dist_min = mean_feature - thr_dist
            thr_dist_max = mean_feature + thr_dist
            aux = feature[
                np.logical_and(feature > thr_dist_min, feature < thr_dist_max)
            ]

            if aux.shape[0] > 10:
                ks.append(test_ks(aux))
            else:
                ks.append(0)

        max_inputs = 0.75
        min_inputs = 10

        # if all:
        # max_inputs = features.shape[1]
        if max_inputs < 1:
            max_inputs = np.ceil(max_inputs * features.shape[1]).astype(int)

        ind = np.argsort(ks)
        A = np.array(ks)[ind]
        A = A[A.shape[0] - max_inputs :]  # Cutoff coeffs

        ncoeff = A.shape[0]
        maxA = A.max()
        nd = 10
        d = (A[nd - 1 :] - A[: -nd + 1]) / maxA * ncoeff / nd
        all_above1 = d[np.nonzero(d >= 1)]
        if all_above1.shape[0] >= 2:
            # temp_bla = smooth(diff(all_above1),3)
            aux2 = np.diff(all_above1)
            temp_bla = np.convolve(aux, np.ones(3) / 3)
            temp_bla = temp_bla[1:-1]
            temp_bla[0] = aux2[0]
            temp_bla[-1] = aux2[-1]
            # ask to be above 1 for 3 consecutive coefficients
            thr_knee_diff = all_above1[np.nonzero(temp_bla[1:] == 1)[:1]] + nd / 2
            inputs = max_inputs - thr_knee_diff + 1
        else:
            inputs = min_inputs

        plot_feature_stats = True
        if plot_feature_stats:
            fig = plt.figure()
            plt.stairs(np.sort(ks))
            plt.plot(
                [len(ks) - inputs + 1, len(ks) - inputs + 1],
                fig.axes[0].get_ylim(),
                "r",
            )
            plt.plot(
                [len(ks) - max_inputs, len(ks) - max_inputs],
                fig.axes[0].get_ylim(),
                "--k",
            )
            plt.ylabel("ks_stat")
            plt.xlabel("# features")
            plt.title(
                f"number of spikes = {number_of_spikes}, inputs_selected = {inputs}"
            )

        if inputs > max_inputs:
            inputs = max_inputs
        elif inputs.shape[0] == 0 or inputs < min_inputs:
            inputs = min_inputs

        coeff = ind[-inputs:]
        # CRATES INPUT MATRIX FOR SPC
        input_for_spc = np.zeros((number_of_spikes, inputs))

        for i in range(number_of_spikes):
            for j in range(inputs):
                input_for_spc[i, j] = features[i, coeff[j]]



__doc__ = """
Auditory Model
==============

References
----------

- https://engineering.purdue.edu/~malcolm/interval/1998-010/
"""

from miv.coding.temporal.lyon import *
from miv.coding.temporal.protocol import *
from miv.coding.temporal.spiker import *



__doc__ = """Sample signal readout from optogenetic neuron."""
__all__ = ["load_data"]

import gzip
import os

import numpy as np

from miv.datasets.utils import get_file
from miv.io import DataManager


def load_data():  # pragma: no cover
    """
    Loads the sample optogenetic experiment data. `Direct Download <https://uofi.box.com/shared/static/9llg11ods9iejdt2omjwjosbsxb5ui10.zip>`_

    Total size: 600.5 MB (compressed)

    File hash: 5deadc1b2a20501b5f6ee8828fa9c85df0b7890bd6ac4eaa8dca768d3b8b5f83

    Notes
    -----
    All experiment are 1 minute long, 30k Hz recording of optogenetic
    neuron cells over 64 channels MEA. Dataset includes 1 spontaneous
    recording and 4 stimulated recording.

    Spontaneous recording is the recording over 1 minute period without
    external stimulation. The purpose was to measure the baseline mean-
    firing rate.

    Stimulation was done by LED light. Over 1 minute (60 seconds) period,
    6 stimulation was done with 10 seconds of intervals. For each stimulation,
    LED light was shined over 1 seconds, followed by remaining 9 seconds
    of rest (without light).

    Containing experiments:

    * experiment0: spontaneous recording
    * experiment1-4: stimulated recordings

    Returns
    -------
    dataset: miv.io.DataManager

    Examples
    --------
        >>> from miv import datasets
        >>> experiments: miv.io.DataManager = datasets.optogenetic.load_data()
        >>> experiments.tree()
        2022-03-10_16-19-09
            0: <miv.io.data.Data object at 0x7fbc30b9a9a0>
               └── Record Node 104/experiment1/recording1
            1: <miv.io.data.Data object at 0x7fbc2ae16700>
               └── Record Node 104/experiment0/recording1
            2: <miv.io.data.Data object at 0x7fbc2ac9e7c0>
               └── Record Node 104/experiment2/recording2
            3: <miv.io.data.Data object at 0x7fbc2ac9edc0>
               └── Record Node 104/experiment3/recording1
            4: <miv.io.data.Data object at 0x7fbc2ac9e160>
               └── Record Node 104/experiment4/recording3

    """

    subdir = "optogenetic"
    base_url = "https://uofi.box.com/shared/static/9llg11ods9iejdt2omjwjosbsxb5ui10.zip"
    file = "2022-03-10_16-19-09.zip"
    file_hash = "5deadc1b2a20501b5f6ee8828fa9c85df0b7890bd6ac4eaa8dca768d3b8b5f83"

    path = get_file(
        file_url=base_url,
        directory=subdir,
        fname=file,
        file_hash=file_hash,
        archive_format="zip",
    )
    experiment = DataManager(path)
    experiment.tree()
    return experiment



from typing import Any, Dict, List, Optional, Sequence, Type, Union

import datetime
import sys
from logging import Logger

import h5py
import numpy as np
from h5py._hl.files import File
from numpy import bytes_


def initialize() -> Dict[str, Any]:
    """Creates an empty data dictionary

    Returns:

        **data** (dict): An empty data dictionary

    """

    data: Dict[str, Any] = {}

    data["_GROUPS_"] = {}
    data["_MAP_DATASETS_TO_COUNTERS_"] = {}
    data["_LIST_OF_COUNTERS_"] = []
    data["_MAP_DATASETS_TO_DATA_TYPES_"] = {}
    data["_GROUP_METADATA_"] = {}

    data["_PROTECTED_NAMES_"] = [
        "_PROTECTED_NAMES_",
        "_GROUPS_",
        "_MAP_DATASETS_TO_COUNTERS_",
        "_MAP_DATASETS_TO_DATA_TYPES_",
        "_LIST_OF_COUNTERS_",
        "_GROUP_METADATA_",
    ]

    return data


def clear_container(container: Dict[str, Any]) -> None:
    """Clears the data from the container dictionary.

    Args:
        **container** (dict): The container to be cleared.

    """

    for key in container.keys():

        if key == "_LIST_OF_COUNTERS_":
            continue

        if isinstance(container[key], list):
            container[key].clear()
        elif isinstance(container[key], np.ndarray):
            container[key] = []
        elif isinstance(container[key], int):
            if key in container["_LIST_OF_COUNTERS_"]:
                container[key] = 0
            else:
                container[key] = -999
        elif isinstance(container[key], float):
            container[key] = np.nan
        elif isinstance(container[key], str):
            container[key] = ""


def create_container(
    data: Dict[str, Any],
) -> Dict[str, Any]:
    """Creates a container dictionary that will be used to collect data and then
    packed into the the master data dictionary.

    Args:
        **data** (dict): Data dictionary that will hold all the data from the container.

    Returns:
        **container** (dict): The new container dictionary with keys and no container information

    """

    container: Dict[str, Any] = {}

    for k in data.keys():
        if k in data["_LIST_OF_COUNTERS_"]:
            container[k] = 0
        else:
            container[k] = data[k].copy()

    return container


def create_group(
    data: Dict[str, Any],
    group_name: str,
    metadata: Dict[str, Union[str, int, float]] = {},
    counter: Optional[str] = None,
    logger: Optional[Logger] = None,
) -> str:
    """Adds a group in the dictionary

    Args:
        **data** (dict): Dictionary to which the group will be added

        **group_name** (string): Name of the group to be added

        **counter** (string): Name of the counter key. None by default

    """

    group_id = group_name.replace("/", "-")
    if logger is not None:
        logger.warning(
            "----------------------------------------------------"
            f"Slashes / are not allowed in group names"
            f"Replacing / with - in group name {group_name}"
            f"The new name will be {group_id}"
            "----------------------------------------------------"
        )
    # Change name of variable, just to keep code more understandable
    counter_id = counter

    # Create a counter_name if the user has not specified one
    if counter_id is None:
        counter_id = f"N_{group_id}"
        if logger is not None:
            logger.warning(
                "----------------------------------------------------"
                f"There is no counter to go with group {group_name}"
                f"Creating a counter called {counter_id}"
                "-----------------------------------------------------"
            )

    # Check for slashes in the counter name. We can't have them.
    if counter_id.find("/") >= 0:
        counter_id = counter_id.replace("/", "-")

    keys = data.keys()

    # Then put the group and any datasets in there next.
    keyfound = False
    for k in keys:
        if group_id == k:
            if logger is not None:
                logger.warning(f"{group_name} is already in the dictionary!")
            keyfound = True
            break

    if not keyfound:

        data["_GROUPS_"][group_id] = []

        data["_GROUPS_"][group_id].append(counter_id)
        counter_path = f"{group_id}/{counter_id}"

        data["_GROUP_METADATA_"][group_id] = metadata
        data["_MAP_DATASETS_TO_COUNTERS_"][group_id] = counter_path
        data["_MAP_DATASETS_TO_DATA_TYPES_"][counter_path] = int

        if counter_path not in data["_LIST_OF_COUNTERS_"]:
            data["_LIST_OF_COUNTERS_"].append(counter_path)

        data[counter_path] = []

    return group_id


def create_dataset(
    data: Dict[str, Any],
    datasets: Union[str, List[str]],
    group: str,
    dtype: Union[Type[int], Type[float], Type[str]] = float,
    logger: Optional[Logger] = None,
) -> int:
    """Adds a dataset to a group in a dictionary. If the group does not exist, it will be created.

    Args:
        **data** (dict): Dictionary that contains the group

        **datasets** (list): Datasets to be added to the group

        **group** (string): Name of group the dataset will be added to.

        **dtype** (type): The data type. float by default.

    Returns:
        **-1**: If the group is None


    """

    if isinstance(datasets, str):
        datasets_: List[str] = [datasets]
    else:
        datasets_ = datasets

    # Check for slashes in the group name. We can't have them.
    for i in range(len(datasets_)):
        dataset_name = datasets_[i]
        if dataset_name.find("/") >= 0:
            new_dataset_name = dataset_name.replace("/", "-")
            datasets_[i] = new_dataset_name
            if logger is not None:
                logger.warning(
                    "----------------------------------------------------"
                    f"Slashes / are not allowed in dataset names"
                    f"Replacing / with - in dataset name {dataset_name}"
                    f"The new name will be {new_dataset_name}"
                    "----------------------------------------------------"
                )

    keys = data.keys()

    if group.find("/") >= 0:
        group = group.replace("/", "-")

    # Put the counter in the dictionary first.
    keyfound = False
    for k in data["_GROUPS_"]:
        if group == k:
            keyfound = True

    if not keyfound:
        counter = f"N_{group}"
        create_group(data, group, counter=counter)
        if logger is not None:
            logger.warning(
                f"Group {group} is not in the dictionary yet!"
                f"Adding it, along with a counter of {counter}"
            )

    # Then put the datasets into the group in there next.
    for dataset in datasets_:
        keyfound = False
        name = f"{group}/{dataset}"
        for k in keys:
            if name == k:
                if logger is not None:
                    logger.warning(f"{name} is already in the dictionary!")
                keyfound = True
        if not keyfound:
            if logger is not None:
                logger.info(
                    f"Adding dataset {dataset} to the dictionary under group {group}."
                )
            data[name] = []
            data["_GROUPS_"][group].append(dataset)

            # Add a counter for this dataset for the group with which it is associated.
            counter = data["_MAP_DATASETS_TO_COUNTERS_"][group]
            # counter_name = "%s/%s" % (group,counter)
            data["_MAP_DATASETS_TO_COUNTERS_"][name] = counter

            data["_MAP_DATASETS_TO_DATA_TYPES_"][name] = dtype

    return 0


def pack(
    data: Dict[str, Any],
    container: Dict[str, Any],
    AUTO_SET_COUNTER: bool = True,
    EMPTY_OUT_CONTAINER: bool = True,
    STRICT_CHECKING: bool = False,
    logger: Optional[Logger] = None,
) -> int:
    """Takes the data from an container and packs it into the data dictionary,
    intelligently, so that it can be stored and extracted efficiently.

    Args:
        **data** (dict): Data dictionary to hold the entire dataset EDIT.

        **container** (dict): container to be packed into data.

        **EMPTY_OUT_CONTAINER** (bool): If this is `True` (default) then empty out the container in preparation
                                for the next iteration. Useful to disable when debugging and inspecting containers.

        **STRICT_CHECKING** (bool): If `True`, then check that all datasets have the same length, otherwise

        **AUTO_SET_COUNTER** (bool): If `True`, update counter value with length of dataset in container

    """

    # Calculate the number of entries for each group and set the
    # value of that counter.
    if AUTO_SET_COUNTER:
        for group in data["_GROUPS_"]:

            datasets = data["_GROUPS_"][group]
            counter = data["_MAP_DATASETS_TO_COUNTERS_"][group]

            # Here we will calculate the values for the counters, based
            # on the size of the datasets
            counter_value = None

            # Loop over the datasets
            for d in datasets:
                full_dataset_name = group + "/" + d
                # Skip any counters
                if counter == full_dataset_name:
                    continue
                else:
                    # Grab the size of the first dataset
                    temp_counter_value = len(container[full_dataset_name])

                    # If we're not STRICT_CHECKING, then use that value for the
                    # counter and break the loop over the datasets, moving on
                    # to the next group.
                    if STRICT_CHECKING is False:
                        container[counter] = temp_counter_value
                        break
                    # Otherwise, we'll check that *all* the datasets have the same
                    # length.
                    else:
                        if counter_value is None:
                            counter_value = temp_counter_value
                            container[counter] = temp_counter_value
                        elif counter_value != temp_counter_value:
                            # In this case, we found two groups of different length!
                            # Print this to help the user identify their error
                            if logger is not None:
                                logger.warning(
                                    f"Two datasets in group {group} have different sizes!"
                                )
                            for tempd in datasets:
                                temp_full_dataset_name = group + "/" + tempd
                                # Don't worry about the dataset
                                if counter == temp_full_dataset_name:
                                    continue

                            # Return a value for the external program to catch.
                            raise RuntimeError(
                                f"Two datasets in group {group} have different sizes!"
                            )

    # Then pack the container into the data
    keys = list(container.keys())
    for key in keys:

        if key in data["_PROTECTED_NAMES_"]:
            continue

        if isinstance(container[key], list):
            value = container[key]
            if len(value) > 0:
                data[key] += value
        else:
            data[key].append(container[key])

    # Clear out the container after it has been packed
    if EMPTY_OUT_CONTAINER:
        clear_container(container)

    return 0


def convert_list_and_key_to_string_data(datalist, key):
    """Converts data dictionary to a string

    Args:
        **datalist** (list): A list to be saved as a string.

    Returns:
        **key** (string): We will assume that this will be unpacked as a dictionary,
                      and this will be the key for the list in that dictionary.

    """

    a = np.string_(key)

    mydataset = []
    b = np.string_("")
    nvals = len(datalist)
    for i, val in enumerate(datalist):
        b += np.string_(val)
        if i < nvals - 1:
            b += np.string_("__:__")
    mydataset.append([a, b])

    return mydataset


def convert_dict_to_string_data(dictionary: Dict[str, str]) -> List[List[bytes_]]:
    """Converts data dictionary to a string

    Args:
        **dictionary** (dict): Dictionary to be converted to a string

    Returns:
        **mydataset** (string): String representation of the dataset

    """

    keys = dictionary.keys()

    mydataset = []
    for i, key in enumerate(keys):
        a = np.string_(key)
        b = np.string_(dictionary[key])
        mydataset.append([a, b])

    return mydataset


def write_metadata(
    filename: str,
    metadata: Dict[str, str] = {},
    write_default_values: bool = True,
    append: bool = True,
) -> File:

    """Writes file metadata in the attributes of an HDF5 file

    Args:
    **filename** (string): Name of output file

    **metadata** (dictionary): Metadata desired by user

    **write_default_values** (boolean): True if user wants to write/update the
                                        default metadata: date, hepfile version,
                                        h5py version, numpy version, and Python
                                        version, false if otherwise.

    **append** (boolean): True if user wants to keep older metadata, false otherwise.

    Returns:
    **hdoutfile** (HDF5): File with new metadata

    """

    hdoutfile = h5py.File(filename, "a")

    non_metadata = ["_NUMBER_OF_CONTAINERS_", "_NUMBER_OF_ENTRIES_"]

    if not append:
        for key in hdoutfile.attr.keys():
            if key not in non_metadata:
                del hdoutfile.attrs[key]

    if write_default_values:
        hdoutfile.attrs["date"] = datetime.datetime.now().isoformat(sep=" ")
        hdoutfile.attrs["numpy_version"] = np.__version__
        hdoutfile.attrs["h5py_version"] = h5py.__version__
        hdoutfile.attrs["python_version"] = sys.version

    for key in metadata:
        hdoutfile.attrs[key] = metadata[key]

    hdoutfile.close()
    return hdoutfile


def write(
    filename: str,
    data: Dict[str, Any],
    comp_type: Optional[str] = None,
    comp_opts: Optional[int] = None,
    logger: Optional[Logger] = None,
) -> File:

    """Writes the selected data to an HDF5 file

    Args:
        **filename** (string): Name of output file

        **data** (dictionary): Data to be written into output file

        **comp_type** (string): Type of compression

    Returns:
        **hdoutfile** (HDF5): File to which the data has been written

    """
    hdoutfile = h5py.File(filename, "w", libver="latest", rdcc_nbytes=1024**3)

    _GROUPS_ = data["_GROUPS_"].keys()

    # Convert this to a 2xN array for writing to the hdf5 file.
    # This gives us one small list of information if we need to pull out
    # small chunks of data
    mydataset = convert_dict_to_string_data(data["_MAP_DATASETS_TO_COUNTERS_"])
    dset = hdoutfile.create_dataset(
        "_MAP_DATASETS_TO_COUNTERS_",
        data=mydataset,
        dtype="S256",
        compression=comp_type,
        compression_opts=comp_opts,
    )

    # Convert this to a 2xN array for writing to the hdf5 file.
    # This has the _GROUPS_ and the datasets in them.
    for group in _GROUPS_:

        hdoutfile.create_group(group)
        hdoutfile[group].attrs["counter"] = np.string_(
            data["_MAP_DATASETS_TO_COUNTERS_"][group]
        )

        metadata = data["_GROUP_METADATA_"][group]
        for key in metadata:
            val = metadata[key]
            if isinstance(val, str):
                hval = np.string_(val)
            else:
                hval = val
            hdoutfile[group].attrs[key] = hval

        datasets = data["_GROUPS_"][group]

        for dataset in datasets:

            name = None
            name = f"{group}/{dataset}"

            x = data[name]

            dataset_dtype = data["_MAP_DATASETS_TO_DATA_TYPES_"][name]

            if isinstance(x, list):
                x = np.asarray(x, dtype=dataset_dtype)

            if logger is not None:
                logger.info(
                    f"Writing dataset {name} to file {name} as type {str(dataset_dtype)}: x.dtype  {x.dtype} data.shape = {x.shape}"
                )

            dset = None
            if dataset_dtype is not str:
                dset = hdoutfile.create_dataset(
                    name,
                    data=x,
                    compression=comp_type,
                    compression_opts=comp_opts,
                    dtype=dataset_dtype,
                    chunks=True,
                )
            else:
                # For writing strings, ensure strings are ascii and not Unicode
                dataset_dtype = h5py.special_dtype(vlen=str)
                longest_word = len(max(x, key=len))
                arr = np.array(x, dtype="S" + str(longest_word))
                dset = hdoutfile.create_dataset(
                    name,
                    data=arr,
                    dtype=dataset_dtype,
                    compression=comp_type,
                    compression_opts=comp_opts,
                )
            dset.attrs["_GROUP_"] = np.string_(group)

    # Get the number of containers
    counters = data["_LIST_OF_COUNTERS_"]
    _NUMBER_OF_CONTAINERS_ = -1
    prevcounter = None
    for i, countername in enumerate(counters):
        ncounter = len(data[countername])
        if logger is not None:
            logger.debug(f"{countername:<32s} has {ncounter:<12d} entries")
        if i > 0 and ncounter != _NUMBER_OF_CONTAINERS_:
            if logger is not None:
                logger.warning(
                    f"{countername} and {prevcounter} have differing numbers of entries!"
                )

        if _NUMBER_OF_CONTAINERS_ < ncounter:
            _NUMBER_OF_CONTAINERS_ = ncounter

        prevcounter = countername

    hdoutfile.attrs["_NUMBER_OF_CONTAINERS_"] = _NUMBER_OF_CONTAINERS_
    hdoutfile.close()

    write_metadata(filename)

    return hdoutfile



__all__ = ["TemporalEncoderProtocol", "SpikerAlgorithmProtocol"]

from typing import Optional, Protocol, Tuple

import numpy as np

from miv.typing import SignalType, SpikestampsType, TimestampsType


class TemporalEncoderProtocol(Protocol):
    def __call__(self, signal: np.ndarray) -> SignalType:  # pragma: no cover
        ...


class SpikerAlgorithmProtocol(Protocol):
    def __init__(self):  # pragma: no cover
        ...

    def __call__(
        self, signal: np.ndarray, sample_rate
    ) -> Tuple[SpikestampsType, TimestampsType]:  # pragma: no cover
        ...

    def save(self) -> None:  # pragma: no cover
        ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['miv',
 'miv.cli',
 'miv.coding',
 'miv.coding.spatial',
 'miv.coding.temporal',
 'miv.core',
 'miv.datasets',
 'miv.io',
 'miv.io.file',
 'miv.io.intan',
 'miv.io.serial',
 'miv.mea',
 'miv.signal',
 'miv.signal.filter',
 'miv.signal.generator',
 'miv.signal.similarity',
 'miv.signal.spike',
 'miv.statistics',
 'miv.visualization']

package_data = \
{'': ['*']}

install_requires = \
['Pillow==9.1.1',
 'PyWavelets>=1.3.0,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'elephant>=0.11.1,<0.12.0',
 'graphviz>=0.20.1,<0.21.0',
 'h5py>=3.7.0,<4.0.0',
 'matplotlib>=3.5.2,<4.0.0',
 'neo>=0.11.0,<0.12.0',
 'numba>=0.56.4,<0.57.0',
 'numpy>=1.23.2,<2.0.0',
 'pandas>=1.4.2,<2.0.0',
 'pyinform>=0.2.0,<0.3.0',
 'pyvis>=0.2.1,<0.4.0',
 'quantities>=0.13.0,<0.14.0',
 'scikit-learn>=1.1.1,<2.0.0',
 'scipy>=1.9.1,<2.0.0',
 'seaborn>=0.11.2,<0.13.0',
 'tqdm>=4.64.0,<5.0.0',
 'viziphant>=0.2.0,<0.3.0']

extras_require = \
{'docs': ['Sphinx>=4.5.0,<5.0.0',
          'pydata-sphinx-theme>=0.9,<0.13',
          'readthedocs-sphinx-search>=0.1.2,<0.2.0',
          'sphinx-autodoc-typehints>=1.19.1,<2.0.0',
          'myst-parser>=0.17.2,<0.18.0',
          'numpydoc>=1.4.0,<2.0.0',
          'sphinx-togglebutton>=0.3.2,<0.4.0',
          'sphinx-copybutton>=0.5.0,<0.6.0',
          'sphinxcontrib-mermaid>=0.7.1,<0.8.0',
          'myst-nb>=0.15.0,<0.16.0'],
 'experiment': ['pyserial>=3.5,<4.0']}

entry_points = \
{'console_scripts': ['convert_open_ephys_to_miv = '
                     'miv.cli.convert_open_ephys_to_miv:main',
                     'miv_extract_spiketrain = '
                     'miv.cli.miv_extract_spiketrain:main']}

setup_kwargs = {
    'name': 'miv-os',
    'version': '0.2.4',
    'description': 'Python software for analysis and computing framework used in MiV project.',
    'long_description': "<div align='center'>\n<h1> MiV-OS: Spike Analysis and Computing Framework </h1>\n\n[![License][badge-LICENSE]][link-LICENSE]\n[![Release pypi][badge-pypi]][link-pypi]\n[![Build Status][badge-CI]][link-CI]\n[![Documentation Status][badge-docs-status]][link-docs-status]\n[![Downloads][badge-pepy-download-count]][link-pepy-download-count]\n[![codecov][badge-codecov]][link-codecov]\n\n</div>\n\n---\n\nPython analysis and computing framework developed for [Mind-in-Vitro(MiV)][link-project-website] project.\n\n## Installation\n[![PyPI version][badge-pypi]][link-pypi]\n\nMiV-OS is compatible with Python 3.8+. The easiest way to install is using python installation package (PIP)\n\n~~~bash\n$ pip install MiV-OS\n~~~\n\n## Documentation\n[![Documentation Status][badge-docs-status]][link-docs-status]\n\nDocumentation of the package is available [here][link-docs-status]\n\n## Contribution\n\nIf you would like to participate, please read our [contribution guideline](CONTRIBUTING.md)\n\nThe development of MiV-OS is lead by the [Gazzola Lab][link-lab-website] at the University of Illinois at Urbana-Champaign.\n\n## List of publications and submissions\n\n## Citation\n\n```\n@misc{MiV-OS,\n  author = {Gazzola Lab},\n  title = {MiV-OS: Analysis and Computation Framework on MiV System and Simulator},\n  year = {2022},\n  publisher = {GitHub},\n  journal = {GitHub repository},\n  howpublished = {\\url{https://github.com/GazzolaLab/MiV-OS}},\n}\n```\n\nWe ask that any publications which use MiV-OS package to cite the following papers:\n\n```\n```\n\n## Developers ✨\n_Names arranged alphabetically_\n- Arman Tekinalp\n- Andrew Dou\n- [Frithjof Gressmann](https://github.com/frthjf)\n- Gaurav Upadhyay\n- [Seung Hyun Kim](https://github.com/skim0119)\n\n[//]: # (Collection of URLs.)\n\n[link-lab-website]: http://mattia-lab.com/\n[link-project-website]: https://mindinvitro.illinois.edu/\n[link-docs-status]: https://miv-os.readthedocs.io/en/latest/?badge=latest\n[link-CI]: https://github.com/GazzolaLab/MiV-OS/actions\n[link-LICENSE]: https://opensource.org/licenses/MIT\n[link-pypi]: https://badge.fury.io/py/MiV-OS\n[link-pepy-download-count]: https://pepy.tech/project/MiV-OS\n[link-codecov]: https://codecov.io/gh/GazzolaLab/MiV-OS\n\n[//]: # (Collection of Badges)\n\n[badge-docs-status]: https://readthedocs.org/projects/miv-os/badge/?version=latest\n[badge-CI]: https://github.com/GazzolaLab/MiV-OS/workflows/CI/badge.svg\n[badge-LICENSE]: https://img.shields.io/badge/License-MIT-yellow.svg\n[badge-pypi]: https://badge.fury.io/py/MiV-OS.svg\n[badge-pepy-download-count]: https://pepy.tech/badge/MiV-OS\n[badge-codecov]: https://codecov.io/gh/GazzolaLab/MiV-OS/branch/main/graph/badge.svg?token=OM5LYWF5KP\n",
    'author': 'GazzolaLab',
    'author_email': 'skim449@illinois.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://mindinvitro.illinois.edu',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import os

import matplotlib.pyplot as plt
import numpy as np

from miv.statistics import burst
from miv.typing import SpikestampsType


def plot_burst(spiketrains: SpikestampsType, min_isi: float, min_len: float):
    """
    Plots burst events across electrodes  to characterize bursting phenomenon on a singl channel

    Parameters
    ----------
    spikes : SpikestampsType
           Single spike-stamps
    min_isi : float
       Minimum Interspike Interval (in seconds) to be considered as bursting [standard = 0.1]
    min_len : float
       Minimum number of simultaneous spikes to be considered as bursting [standard = 10]

    Returns
    -------
    figure, axes
       matplot figure with bursts plotted for all electrodes
    """

    fig, ax = plt.subplots()
    start_time = []
    burst_duration = []
    burst_len = []
    burst_rate = []
    for i in np.arange(len(spiketrains)):
        start_time, burst_duration, burst_len, burst_rate = burst(
            spiketrains, i, min_isi, min_len
        )
        a = np.column_stack((start_time, burst_duration))
        ax.broken_barh(a, (i + 1, 0.5), facecolors="tab:orange")

    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Electrode")

    return fig, ax



__doc__ = """Multi-channel signal plotting for MEA channels"""
__all__ = ["multi_channel_signal_plot"]

from typing import Any, List, Optional

import matplotlib
import matplotlib.animation as manimation
import numpy as np
from matplotlib import pyplot as plt
from tqdm import tqdm

from miv.mea.protocol import MEAGeometryProtocol
from miv.typing import SignalType

# matplotlib.use("Agg")  # Must be before importing matplotlib.pyplot or pylab!


def multi_channel_signal_plot(
    signal_list: SignalType,
    mea_geometry: MEAGeometryProtocol,
    start_step: int,
    end_step: int,
    n_steps_in_window: int,
    rendering_fps: int,
    video_name: str,
    max_subplot_in_x: int = 8,
    max_subplot_in_y: int = 8,
    dpi: int = 100,
):
    """
    Plotting recorded neuron signals from each channel of MEA. Subplots for each channel are aligned with position of
    electrical probes on MEA.

    Parameters
    ----------
    signal_list : list
        Contains list  2D numpy.ndarray
        List of channel recordings in time.
    mea_geometry : list
        Contains list of tuples.
        Each tuple contains, channel id, channel y position id and channel x position id on MEA grid.
    start_step : int
        Start step for plotting.
    end_step : int
        End step for plotting.
    n_steps_in_window : int
        Window length for plotting channel signal.
    rendering_fps : int
        Video frame rate
    video_name : str
        Video name
    max_subplot_in_x : int
        (default=8)
    max_subplot_in_y : int
        (default=8)
    dpi : int
        (default=100)

    Returns
    -------

    """
    total_steps = end_step - start_step - n_steps_in_window

    channel_id = []
    xid = []
    yid = []
    for i, channel_info in enumerate(mea_geometry):
        channel_id.append(channel_info[0])
        yid.append(channel_info[1])
        xid.append(channel_info[2])
    n_channels = len(channel_id)

    FFMpegWriter = manimation.writers["ffmpeg"]
    metadata = dict(title="Movie Test", artist="Matplotlib", comment="Movie support!")
    writer = FFMpegWriter(fps=rendering_fps, metadata=metadata)
    fig = plt.figure(2, figsize=(20, 12), frameon=True, dpi=dpi)
    plt.rcParams.update({"font.size": 10})
    axs = []

    for y, x in zip(yid, xid):
        axs.append(plt.subplot2grid((max_subplot_in_y, max_subplot_in_x), (y, x)))

    signal_line_list = [None for _ in range(n_channels)]

    for signal_id in range(n_channels):
        signal = signal_list[:, signal_id]

        # x_value = timestamps[start_step : start_step + n_steps_in_window]
        y_value = signal[start_step : start_step + n_steps_in_window]
        signal_line_list[signal_id] = axs[signal_id].plot(y_value, "-", linewidth=3)[0]

        y_min = np.min(signal[start_step:end_step])
        y_max = np.max(signal[start_step:end_step])

        axs[signal_id].set_ylim(y_min, y_max)

    plt.tight_layout()
    fig.align_ylabels()

    with writer.saving(fig, video_name, dpi):
        for step in tqdm(range(total_steps)):
            current_step = start_step + step
            for signal_id in range(n_channels):
                signal = signal_list[:, signal_id]
                # x_value = signal[current_step : current_step + n_steps_in_window, 0]
                y_value = signal[current_step : current_step + n_steps_in_window]

                # signal_line_list[signal_id].set_xdata(x_value)
                signal_line_list[signal_id].set_ydata(y_value)

                # X limits should move together with window
                # axs[signal_id].set_xlim(x_value[0], x_value[-1])

            writer.grab_frame()

    # Be a good boy and close figures
    # https://stackoverflow.com/a/37451036
    # plt.close(fig) alone does not suffice
    # See https://github.com/matplotlib/matplotlib/issues/8560/
    plt.close(plt.gcf())



from typing import Any, Dict, List, Optional, Tuple, Union

from logging import Logger

import h5py
import numpy as np
from h5py._hl.dataset import Dataset
from numpy import int64, ndarray


def read(
    filename: str,
    groups: Optional[Union[str, List[str]]] = None,
    subset: Optional[Union[int, List[int], Tuple[int, int]]] = None,
    logger: Optional[Logger] = None,
) -> Tuple[Dict[str, Any], Dict[str, None]]:
    """
    Reads all, or a subset of the data, from the HDF5 file to fill a data dictionary.
    Returns an empty dictionary to be filled later with data from individual containers.

    Args:
        **filename** (string): Name of the input file

        **groups** (list): Dataset groups to be read from input file.

        **subset** (int): Number of containers to be read from input file

        **logger** (logging.Logger): optional logger object

    Returns:
        **data** (dict): Selected data from HDF5 file

        **container** (dict): An empty container dictionary to be filled by data from select containers

    """

    # Open the HDF5 file
    infile = None
    infile = h5py.File(filename, "r")

    # Create the initial data and container dictionary to hold the data
    data: Dict[str, Any] = {}
    container: Dict[str, Any] = {}

    data["_MAP_DATASETS_TO_COUNTERS_"] = {}
    data["_MAP_DATASETS_TO_INDEX_"] = {}
    data["_LIST_OF_COUNTERS_"] = []
    data["_LIST_OF_DATASETS_"] = []
    data["_GROUPS_"] = []
    data["_MAP_DATASETS_TO_GROUPS_"] = {}

    # Get the number of containers
    data["_NUMBER_OF_CONTAINERS_"] = infile.attrs["_NUMBER_OF_CONTAINERS_"]

    ncontainers = data["_NUMBER_OF_CONTAINERS_"]

    # Determine if only a subset of the data should be read
    subset_: Union[None, List[int]] = None
    if subset is not None:

        try:
            subset_ = validate_subset(subset, ncontainers)
        except RuntimeError:
            infile.close()
            return data, container

        data["_NUMBER_OF_CONTAINERS_"] = subset_[1] - subset_[0]
        ncontainers = data["_NUMBER_OF_CONTAINERS_"]

        if logger is not None:
            logger.info(
                "Will read in a subset of the file!"
                f"From container {subset_[0]} (inclusive) through container {subset_[1]-1} (inclusive)"
                f"Container {subset_[1]} is not read in"
                f"Reading in {ncontainers} containers\n"
            )

    # Get the datasets and counters
    dc = infile["_MAP_DATASETS_TO_COUNTERS_"]
    for vals in dc:

        # The decode is there because vals were stored as numpy.bytes
        counter = vals[1].decode()
        index_name = f"{counter}_INDEX"
        data["_MAP_DATASETS_TO_COUNTERS_"][vals[0].decode()] = counter
        data["_MAP_DATASETS_TO_INDEX_"][vals[0].decode()] = index_name
        data["_LIST_OF_COUNTERS_"].append(vals[1].decode())
        data["_LIST_OF_DATASETS_"].append(vals[0].decode())
        data["_LIST_OF_DATASETS_"].append(vals[1].decode())  # Get the counters as well

    # We may have added some counters and datasets multiple times.
    # So just to be sure, only keep the unique values
    data["_LIST_OF_COUNTERS_"] = np.unique(data["_LIST_OF_COUNTERS_"]).tolist()
    data["_LIST_OF_DATASETS_"] = np.unique(data["_LIST_OF_DATASETS_"]).tolist()

    # Get the list of datasets and groups
    all_datasets = data["_LIST_OF_DATASETS_"]

    all_datasets = select_datasets(all_datasets, groups, logger=logger)

    # Pull out the counters and build the indices
    # We will need to keep track of the indices in the entire file
    # This way, if the user specifies a subset of the data, we have the full
    # indices already calculated
    full_file_indices = {}

    for counter_name in data["_LIST_OF_COUNTERS_"]:

        full_file_counters = infile[counter_name]
        full_file_index = calculate_index_from_counters(full_file_counters)

        # If we passed in subset, grab that slice of the data from the file
        if subset_ is not None:
            # Add 1 to the high range of subset when we pull out the counters
            # and index because in order to get all of the entries for the last entry.
            data[counter_name] = infile[counter_name][subset_[0] : subset_[1] + 1]
            index: np.ndarray = full_file_index[subset_[0] : subset_[1] + 1]
        else:
            data[counter_name] = infile[counter_name][:]
            index = full_file_index

        subset_index = index
        # If the file is *not* empty....
        # Just to make sure the "local" index of the data dictionary starts at 0
        if len(index) > 0:
            subset_index = index - index[0]

        index_name = f"{counter_name}_INDEX"

        data[index_name] = subset_index
        full_file_indices[index_name] = index

    # Loop over the all_datasets we want and pull out the data.
    for name in all_datasets:

        # If this is a counter, we're going to have to grab the indices
        # differently than for a "normal" dataset
        IS_COUNTER = True
        index_name_: Union[None, str] = None
        if name not in data["_LIST_OF_COUNTERS_"]:
            index_name_ = data["_MAP_DATASETS_TO_INDEX_"][name]
            IS_COUNTER = False  # We will use different indices for the counters

        dataset = infile[name]

        # This will ignore the groups
        if isinstance(dataset, h5py.Dataset):
            dataset_name = name
            group_name = dataset.attrs.get("_GROUP_", None)

            if subset_ is not None:
                if IS_COUNTER:
                    # If this is a counter, then the subset indices
                    # map on to the same locations for any counters
                    lo = subset_[0]
                    hi = subset_[1]
                else:
                    if index_name_ is not None:
                        lo = full_file_indices[index_name_][0]
                        hi = full_file_indices[index_name_][-1]
                    else:
                        raise RuntimeError("Unknown index")
                data[dataset_name] = dataset[lo:hi]
            else:
                data[dataset_name] = dataset[:]
            if group_name not in data["_GROUPS_"]:
                data["_GROUPS_"].append(group_name)
            if dataset_name not in data["_MAP_DATASETS_TO_GROUPS_"]:
                data["_MAP_DATASETS_TO_GROUPS_"][dataset_name] = group_name

            container[
                dataset_name
            ] = None  # This will be filled for individual container

    infile.close()
    return data, container


def select_datasets(
    datasets: List[str],
    groups: Optional[Union[str, List[str]]] = None,
    logger: Optional[Logger] = None,
) -> List[str]:

    # Only keep select data from file, if we have specified datasets
    if groups is not None:

        if isinstance(groups, str):
            groups = list(groups)

        # Count backwards because we'll be removing stuff as we go.
        i = len(datasets) - 1
        while i >= 0:
            entry = datasets[i]

            is_dropped = True
            # This is looking to see if the string is anywhere in the name
            # of the dataset
            for group in groups:
                if group in entry:
                    is_dropped = False
                    break

            if is_dropped:
                if logger is not None:
                    logger.info(f"Not reading out {entry} from the file....")
                datasets.remove(entry)

            i -= 1

        if logger is not None:
            logger.debug(
                f"After only selecting certain datasets ----- " f"datasets: {datasets}"
            )

    return datasets


def validate_subset(
    subset: Union[int, List[int], Tuple[int, int]],
    ncontainers: int,
    logger: Optional[Logger] = None,
) -> List[int]:

    if isinstance(subset, tuple):
        subset_ = list(subset)

    elif isinstance(subset, list):
        subset_ = subset

    elif isinstance(subset, int):
        if logger is not None:
            logger.warning(
                "Single subset value of {subset} being interpreted as a high range"
                f"subset being set to a range of (0,{subset})\n"
            )
        subset_ = [0, subset]
    else:
        raise RuntimeError(f"Unsupported type of subset argument: {subset}")

    # If the user has specified `subset` incorrectly, then let's return
    # an empty data and container
    if subset_[1] - subset_[0] <= 0:
        if logger is not None:
            logger.warning(
                "The range in subset is either 0 or negative!"
                f"{subset_[1]} - {subset_[0]} = {subset_[1] - subset_[0]}"
                "Returning an empty data and container dictionary!\n"
            )
        raise RuntimeError("Range in subset is 0 or negative")

    if subset_[0] > ncontainers:
        if logger is not None:
            logger.error(
                "Range for subset starts greater than number of containers in file!"
                f"{subset_[0]} > {ncontainers}"
            )
        raise RuntimeError(
            "Range for subset starts greater than number of containers in file!"
        )

    if subset_[1] > ncontainers:
        if logger is not None:
            logger.warning(
                "Range for subset is greater than number of containers in file!"
                f"{subset_[1]} > {ncontainers}"
                f"High range of subset will be set to {ncontainers}\n"
            )
        subset_[1] = ncontainers

    if subset_[1] <= subset_[0]:
        if logger is not None:
            logger.error(
                "Will not be reading anything in!"
                f"High range of {subset_[1]} is less than or equal to low range of {subset_[0]}"
            )
        raise RuntimeError(
            f"High range of {subset_[1]} is less than or equal to low range of {subset_[0]}"
        )

    return subset_


def calculate_index_from_counters(counters: Dataset) -> ndarray:
    index = np.add.accumulate(counters) - counters
    return index


def unpack(
    container: Dict[str, Any],
    data: Dict[str, Any],
    n: int = 0,
) -> None:

    """Fills the container dictionary with selected rows from the data dictionary.

    Args:

        **container** (dict): container dictionary to be filled

        **data** (dict): Data dictionary used to fill the container dictionary

    **n** (integer): 0 by default. Which entry should be pulled out of the data
                     dictionary and inserted into the container dictionary.

    """

    keys = container.keys()

    for key in keys:

        # if "num" in key:
        if key in data["_LIST_OF_COUNTERS_"]:
            container[key] = data[key][n]

        elif "INDEX" not in key:
            indexkey = data["_MAP_DATASETS_TO_INDEX_"][key]
            numkey = data["_MAP_DATASETS_TO_COUNTERS_"][key]

            if len(data[indexkey]) > 0:
                index = data[indexkey][n]

            if len(data[numkey]) > 0:
                nobjs = data[numkey][n]
                container[key] = data[key][index : index + nobjs]


def get_ncontainers_in_file(
    filename: str, logger: Optional[Logger] = None
) -> Union[None, int64]:

    """Get the number of containers in the file."""

    with h5py.File(filename, "r+") as f:
        a = f.attrs

        if a.__contains__("_NUMBER_OF_CONTAINERS_"):
            _NUMBER_OF_CONTAINERS_ = a.get("_NUMBER_OF_CONTAINERS_")
            f.close()
            return _NUMBER_OF_CONTAINERS_
        else:
            if logger is not None:
                logger.warning(
                    '\nFile does not contain the attribute, "_NUMBER_OF_CONTAINERS_"\n'
                )
            f.close()
            return None


def get_ncontainers_in_data(data, logger=None) -> Union[None, int64]:

    """Get the number of containers in the data dictionary.

    This is useful in case you've only pulled out subsets of the data

    """

    if not isinstance(data, dict):
        if logger is not None:
            logger.warning(f"{data} is not a dictionary!\n")
        return None

    if "_NUMBER_OF_CONTAINERS_" in list(data.keys()):
        _NUMBER_OF_CONTAINERS_ = data["_NUMBER_OF_CONTAINERS_"]
        return _NUMBER_OF_CONTAINERS_
    else:
        if logger is not None:
            logger.warning(
                '\ndata dictionary does not contain the key, "_NUMBER_OF_CONTAINERS_"\n'
            )
        return None


def get_file_metadata(filename: str) -> Union[None, Dict[str, Any]]:

    """Get the file metadata and return it as a dictionary"""

    f = h5py.File(filename, "r+")

    a = f.attrs

    if len(a) < 1:
        f.close()
        return None

    metadata = {}

    for key in a.keys():
        metadata[key] = a[key]

    f.close()

    return metadata


def print_file_metadata(filename: str):

    """Pretty print the file metadata"""

    metadata = get_file_metadata(filename)

    if metadata is None:
        return None

    output = ""

    keys = list(metadata.keys())

    first_keys_to_print = ["date", "_NUMBER_OF_CONTAINERS_"]

    keys_already_printed = []

    # Print the basics first
    for key in first_keys_to_print:
        if key in keys:
            val = metadata[key]
            output += f"{key:<20s} : {val}\n"
            keys_already_printed.append(key)

    # Print the versions next
    for key in keys:
        if key in keys_already_printed:
            continue

        if key.find("version") >= 0:
            val = metadata[key]
            output += f"{key:<20s} : {val}\n"
            keys_already_printed.append(key)

    # Print the read of the metadata
    for key in keys:
        if key in keys_already_printed:
            continue

        val = metadata[key]
        output += f"{key:<20s} : {val}\n"
        keys_already_printed.append(key)

    return output
